#!/usr/bin/env python3
# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Fixes identifiers that would cause compiler errors in generated C++ code."""

from typing import Set

# Set of words that can't be used as identifiers in the generated code. Many of
# these are valid identifiers in proto syntax, but they need special handling in
# the generated C++ code.
#
# Note: This is primarily used for "if x in y" operations, hence the use of a
# set rather than a list.
PW_PROTO_CODEGEN_RESERVED_WORDS: Set[str] = {
    # Identifiers that conflict with the codegen internals when used in certain
    # contexts:
    "Fields",
    "Message",
    # C++20 keywords (https://en.cppreference.com/w/cpp/keyword):
    "alignas",
    "alignof",
    "and",
    "and_eq",
    "asm",
    "atomic_cancel",
    "atomic_commit",
    "atomic_noexcept",
    "auto",
    "bitand",
    "bitor",
    "bool",
    "break",
    "case",
    "catch",
    "char",
    "char8_t",
    "char16_t",
    "char32_t",
    "class",
    "compl",
    "concept",
    "const",
    "consteval",
    "constexpr",
    "constinit",
    "const_cast",
    "continue",
    "co_await",
    "co_return",
    "co_yield",
    "decltype",
    "default",
    "delete",
    "do",
    "double",
    "dynamic_cast",
    "else",
    "enum",
    "explicit",
    "export",
    "extern",
    "false",
    "float",
    "for",
    "friend",
    "goto",
    "if",
    "inline",
    "int",
    "long",
    "mutable",
    "namespace",
    "new",
    "noexcept",
    "not",
    "not_eq",
    "nullptr",
    "operator",
    "or",
    "or_eq",
    "private",
    "protected",
    "public",
    "reflexpr",
    "register",
    "reinterpret_cast",
    "requires",
    "return",
    "short",
    "signed",
    "sizeof",
    "static",
    "static_assert",
    "static_cast",
    "struct",
    "switch",
    "synchronized",
    "template",
    "this",
    "thread_local",
    "throw",
    "true",
    "try",
    "typedef",
    "typeid",
    "typename",
    "union",
    "unsigned",
    "using",
    "virtual",
    "void",
    "volatile",
    "wchar_t",
    "while",
    "xor",
    "xor_eq",
    # C++20 macros (https://en.cppreference.com/w/cpp/symbol_index/macro),
    # excluding the following:
    # - Function-like macros, which have unambiguous syntax and thus won't
    #   conflict with generated symbols.
    # - Macros that couldn't be made valid by appending underscores, namely
    #   those containing "__" or starting with "_[A-Z]". C++ reserves all such
    #   identifiers for the compiler, and appending underscores wouldn't change
    #   that.
    "ATOMIC_BOOL_LOCK_FREE",
    "ATOMIC_CHAR_LOCK_FREE",
    "ATOMIC_CHAR16_T_LOCK_FREE",
    "ATOMIC_CHAR32_T_LOCK_FREE",
    "ATOMIC_CHAR8_T_LOCK_FREE",
    "ATOMIC_FLAG_INIT",
    "ATOMIC_INT_LOCK_FREE",
    "ATOMIC_LLONG_LOCK_FREE",
    "ATOMIC_LONG_LOCK_FREE",
    "ATOMIC_POINTER_LOCK_FREE",
    "ATOMIC_SHORT_LOCK_FREE",
    "ATOMIC_WCHAR_T_LOCK_FREE",
    "BUFSIZ",
    "CHAR_BIT",
    "CHAR_MAX",
    "CHAR_MIN",
    "CLOCKS_PER_SEC",
    "DBL_DECIMAL_DIG",
    "DBL_DIG",
    "DBL_EPSILON",
    "DBL_HAS_SUBNORM",
    "DBL_MANT_DIG",
    "DBL_MAX",
    "DBL_MAX_10_EXP",
    "DBL_MAX_EXP",
    "DBL_MIN",
    "DBL_MIN_10_EXP",
    "DBL_MIN_EXP",
    "DBL_TRUE_MIN",
    "DECIMAL_DIG",
    "E2BIG",
    "EACCES",
    "EADDRINUSE",
    "EADDRNOTAVAIL",
    "EAFNOSUPPORT",
    "EAGAIN",
    "EALREADY",
    "EBADF",
    "EBADMSG",
    "EBUSY",
    "ECANCELED",
    "ECHILD",
    "ECONNABORTED",
    "ECONNREFUSED",
    "ECONNRESET",
    "EDEADLK",
    "EDESTADDRREQ",
    "EDOM",
    "EEXIST",
    "EFAULT",
    "EFBIG",
    "EHOSTUNREACH",
    "EIDRM",
    "EILSEQ",
    "EINPROGRESS",
    "EINTR",
    "EINVAL",
    "EIO",
    "EISCONN",
    "EISDIR",
    "ELOOP",
    "EMFILE",
    "EMLINK",
    "EMSGSIZE",
    "ENAMETOOLONG",
    "ENETDOWN",
    "ENETRESET",
    "ENETUNREACH",
    "ENFILE",
    "ENOBUFS",
    "ENODATA",
    "ENODEV",
    "ENOENT",
    "ENOEXEC",
    "ENOLCK",
    "ENOLINK",
    "ENOMEM",
    "ENOMSG",
    "ENOPROTOOPT",
    "ENOSPC",
    "ENOSR",
    "ENOSTR",
    "ENOSYS",
    "ENOTCONN",
    "ENOTDIR",
    "ENOTEMPTY",
    "ENOTRECOVERABLE",
    "ENOTSOCK",
    "ENOTSUP",
    "ENOTTY",
    "ENXIO",
    "EOF",
    "EOPNOTSUPP",
    "EOVERFLOW",
    "EOWNERDEAD",
    "EPERM",
    "EPIPE",
    "EPROTO",
    "EPROTONOSUPPORT",
    "EPROTOTYPE",
    "ERANGE",
    "EROFS",
    "errno",
    "ESPIPE",
    "ESRCH",
    "ETIME",
    "ETIMEDOUT",
    "ETXTBSY",
    "EWOULDBLOCK",
    "EXDEV",
    "EXIT_FAILURE",
    "EXIT_SUCCESS",
    "FE_ALL_EXCEPT",
    "FE_DFL_ENV",
    "FE_DIVBYZERO",
    "FE_DOWNWARD",
    "FE_INEXACT",
    "FE_INVALID",
    "FE_OVERFLOW",
    "FE_TONEAREST",
    "FE_TOWARDZERO",
    "FE_UNDERFLOW",
    "FE_UPWARD",
    "FILENAME_MAX",
    "FLT_DECIMAL_DIG",
    "FLT_DIG",
    "FLT_EPSILON",
    "FLT_EVAL_METHOD",
    "FLT_HAS_SUBNORM",
    "FLT_MANT_DIG",
    "FLT_MAX",
    "FLT_MAX_10_EXP",
    "FLT_MAX_EXP",
    "FLT_MIN",
    "FLT_MIN_10_EXP",
    "FLT_MIN_EXP",
    "FLT_RADIX",
    "FLT_ROUNDS",
    "FLT_TRUE_MIN",
    "FOPEN_MAX",
    "FP_FAST_FMA",
    "FP_FAST_FMAF",
    "FP_FAST_FMAL",
    "FP_ILOGB0",
    "FP_ILOGBNAN",
    "FP_SUBNORMAL",
    "FP_ZERO",
    "FP_INFINITE",
    "FP_NAN",
    "FP_NORMAL",
    "HUGE_VAL",
    "HUGE_VALF",
    "HUGE_VALL",
    "INFINITY",
    "INT_FAST16_MAX",
    "INT_FAST16_MIN",
    "INT_FAST32_MAX",
    "INT_FAST32_MIN",
    "INT_FAST64_MAX",
    "INT_FAST64_MIN",
    "INT_FAST8_MAX",
    "INT_FAST8_MIN",
    "INT_LEAST16_MAX",
    "INT_LEAST16_MIN",
    "INT_LEAST32_MAX",
    "INT_LEAST32_MIN",
    "INT_LEAST64_MAX",
    "INT_LEAST64_MIN",
    "INT_LEAST8_MAX",
    "INT_LEAST8_MIN",
    "INT_MAX",
    "INT_MIN",
    "INT16_MAX",
    "INT16_MIN",
    "INT32_MAX",
    "INT32_MIN",
    "INT64_MAX",
    "INT64_MIN",
    "INT8_MAX",
    "INT8_MIN",
    "INTMAX_MAX",
    "INTMAX_MIN",
    "INTPTR_MAX",
    "INTPTR_MIN",
    "L_tmpnam",
    "LC_ALL",
    "LC_COLLATE",
    "LC_CTYPE",
    "LC_MONETARY",
    "LC_NUMERIC",
    "LC_TIME",
    "LDBL_DECIMAL_DIG",
    "LDBL_DIG",
    "LDBL_EPSILON",
    "LDBL_HAS_SUBNORM",
    "LDBL_MANT_DIG",
    "LDBL_MAX",
    "LDBL_MAX_10_EXP",
    "LDBL_MAX_EXP",
    "LDBL_MIN",
    "LDBL_MIN_10_EXP",
    "LDBL_MIN_EXP",
    "LDBL_TRUE_MIN",
    "LLONG_MAX",
    "LLONG_MIN",
    "LONG_MAX",
    "LONG_MIN",
    "MATH_ERREXCEPT",
    "math_errhandling",
    "MATH_ERRNO",
    "MB_CUR_MAX",
    "MB_LEN_MAX",
    "NAN",
    "NULL",
    "ONCE_FLAG_INIT",
    "PRId16",
    "PRId32",
    "PRId64",
    "PRId8",
    "PRIdFAST16",
    "PRIdFAST32",
    "PRIdFAST64",
    "PRIdFAST8",
    "PRIdLEAST16",
    "PRIdLEAST32",
    "PRIdLEAST64",
    "PRIdLEAST8",
    "PRIdMAX",
    "PRIdPTR",
    "PRIi16",
    "PRIi32",
    "PRIi64",
    "PRIi8",
    "PRIiFAST16",
    "PRIiFAST32",
    "PRIiFAST64",
    "PRIiFAST8",
    "PRIiLEAST16",
    "PRIiLEAST32",
    "PRIiLEAST64",
    "PRIiLEAST8",
    "PRIiMAX",
    "PRIiPTR",
    "PRIo16",
    "PRIo32",
    "PRIo64",
    "PRIo8",
    "PRIoFAST16",
    "PRIoFAST32",
    "PRIoFAST64",
    "PRIoFAST8",
    "PRIoLEAST16",
    "PRIoLEAST32",
    "PRIoLEAST64",
    "PRIoLEAST8",
    "PRIoMAX",
    "PRIoPTR",
    "PRIu16",
    "PRIu32",
    "PRIu64",
    "PRIu8",
    "PRIuFAST16",
    "PRIuFAST32",
    "PRIuFAST64",
    "PRIuFAST8",
    "PRIuLEAST16",
    "PRIuLEAST32",
    "PRIuLEAST64",
    "PRIuLEAST8",
    "PRIuMAX",
    "PRIuPTR",
    "PRIx16",
    "PRIX16",
    "PRIx32",
    "PRIX32",
    "PRIx64",
    "PRIX64",
    "PRIx8",
    "PRIX8",
    "PRIxFAST16",
    "PRIXFAST16",
    "PRIxFAST32",
    "PRIXFAST32",
    "PRIxFAST64",
    "PRIXFAST64",
    "PRIxFAST8",
    "PRIXFAST8",
    "PRIxLEAST16",
    "PRIXLEAST16",
    "PRIxLEAST32",
    "PRIXLEAST32",
    "PRIxLEAST64",
    "PRIXLEAST64",
    "PRIxLEAST8",
    "PRIXLEAST8",
    "PRIxMAX",
    "PRIXMAX",
    "PRIxPTR",
    "PRIXPTR",
    "PTRDIFF_MAX",
    "PTRDIFF_MIN",
    "RAND_MAX",
    "SCHAR_MAX",
    "SCHAR_MIN",
    "SCNd16",
    "SCNd32",
    "SCNd64",
    "SCNd8",
    "SCNdFAST16",
    "SCNdFAST32",
    "SCNdFAST64",
    "SCNdFAST8",
    "SCNdLEAST16",
    "SCNdLEAST32",
    "SCNdLEAST64",
    "SCNdLEAST8",
    "SCNdMAX",
    "SCNdPTR",
    "SCNi16",
    "SCNi32",
    "SCNi64",
    "SCNi8",
    "SCNiFAST16",
    "SCNiFAST32",
    "SCNiFAST64",
    "SCNiFAST8",
    "SCNiLEAST16",
    "SCNiLEAST32",
    "SCNiLEAST64",
    "SCNiLEAST8",
    "SCNiMAX",
    "SCNiPTR",
    "SCNo16",
    "SCNo32",
    "SCNo64",
    "SCNo8",
    "SCNoFAST16",
    "SCNoFAST32",
    "SCNoFAST64",
    "SCNoFAST8",
    "SCNoLEAST16",
    "SCNoLEAST32",
    "SCNoLEAST64",
    "SCNoLEAST8",
    "SCNoMAX",
    "SCNoPTR",
    "SCNu16",
    "SCNu32",
    "SCNu64",
    "SCNu8",
    "SCNuFAST16",
    "SCNuFAST32",
    "SCNuFAST64",
    "SCNuFAST8",
    "SCNuLEAST16",
    "SCNuLEAST32",
    "SCNuLEAST64",
    "SCNuLEAST8",
    "SCNuMAX",
    "SCNuPTR",
    "SCNx16",
    "SCNx32",
    "SCNx64",
    "SCNx8",
    "SCNxFAST16",
    "SCNxFAST32",
    "SCNxFAST64",
    "SCNxFAST8",
    "SCNxLEAST16",
    "SCNxLEAST32",
    "SCNxLEAST64",
    "SCNxLEAST8",
    "SCNxMAX",
    "SCNxPTR",
    "SEEK_CUR",
    "SEEK_END",
    "SEEK_SET",
    "SHRT_MAX",
    "SHRT_MIN",
    "SIG_ATOMIC_MAX",
    "SIG_ATOMIC_MIN",
    "SIG_DFL",
    "SIG_ERR",
    "SIG_IGN",
    "SIGABRT",
    "SIGFPE",
    "SIGILL",
    "SIGINT",
    "SIGSEGV",
    "SIGTERM",
    "SIZE_MAX",
    "stderr",
    "stdin",
    "stdout",
    "TIME_UTC",
    "TMP_MAX",
    "UCHAR_MAX",
    "UINT_FAST16_MAX",
    "UINT_FAST32_MAX",
    "UINT_FAST64_MAX",
    "UINT_FAST8_MAX",
    "UINT_LEAST16_MAX",
    "UINT_LEAST32_MAX",
    "UINT_LEAST64_MAX",
    "UINT_LEAST8_MAX",
    "UINT_MAX",
    "UINT16_MAX",
    "UINT32_MAX",
    "UINT64_MAX",
    "UINT8_MAX",
    "UINTMAX_MAX",
    "UINTPTR_MAX",
    "ULLONG_MAX",
    "ULONG_MAX",
    "USHRT_MAX",
    "WCHAR_MAX",
    "WCHAR_MIN",
    "WEOF",
    "WINT_MAX",
    "WINT_MIN",
}


def _transform_invalid_identifier(invalid_identifier: str) -> str:
    """Applies a transformation to an invalid C++ identifier to make it valid.

    Currently, this simply appends an underscore. This addresses the vast
    majority of realistic cases, but there are some caveats; see
    `fix_cc_identifier` function documentation for details.
    """
    return f"{invalid_identifier}_"


def fix_cc_identifier(proto_identifier: str) -> str:
    """Returns an adjusted form of the identifier for use in generated C++ code.

    If the given identifier is already valid for use in the generated C++ code,
    it will be returned as-is. If the identifier is a C++ keyword or a
    preprocessor macro from the standard library, the returned identifier will
    be modified slightly in order to avoid compiler errors.

    Currently, this simply appends an underscore if necessary. This handles the
    vast majority of realistic cases, though it doesn't attempt to fix
    identifiers that the C++ spec reserves for the compiler's use.

    For reference, C++ reserves two categories of identifiers for the compiler:
    - Any identifier that contains the substring "__" anywhere in it.
    - Any identifier with an underscore for the first character and a capital
      letter for the second character.
    """
    return (
        _transform_invalid_identifier(proto_identifier)  #
        if proto_identifier in PW_PROTO_CODEGEN_RESERVED_WORDS  #
        else proto_identifier
    )


def fix_cc_enum_value_name(proto_enum_entry: str) -> str:
    """Returns an adjusted form of the enum-value name for use in generated C++.

    Generates an UPPER_SNAKE_CASE variant of the given enum-value name and then
    checks it for collisions with C++ keywords and standard-library macros.
    Returns a potentially modified version of the input in order to fix
    collisions if any are found.

    Note that, although the code generation also creates enum-value aliases in
    kHungarianNotationPascalCase, symbols of that form never conflict with
    keywords or standard-library macros in C++20. Therefore, only the
    UPPER_SNAKE_CASE versions need to be checked for conflicts.

    See `fix_cc_identifier` for further details.
    """
    upper_snake_case = proto_enum_entry.upper()
    return (
        _transform_invalid_identifier(proto_enum_entry)  #
        if upper_snake_case in PW_PROTO_CODEGEN_RESERVED_WORDS  #
        else proto_enum_entry
    )



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Install and check status of teensy-core."""

import json
import logging
import re
import subprocess
import tempfile
from pathlib import Path
from typing import Sequence

from pw_arduino_build import core_installer

import pw_package.package_manager

_LOG: logging.Logger = logging.getLogger(__name__)


class ArduinoCore(pw_package.package_manager.Package):
    """Install and check status of arduino cores."""

    def __init__(self, core_name, *args, **kwargs):
        super().__init__(*args, name=core_name, **kwargs)

    def status(self, path: Path) -> bool:
        return (path / 'hardware').is_dir()

    def populate_download_cache_from_cipd(self, path: Path) -> None:
        """Check for arduino core availability in pigweed_internal cipd."""
        package_path = path.parent.resolve()
        core_name = self.name
        core_cache_path = package_path / ".cache" / core_name
        core_cache_path.mkdir(parents=True, exist_ok=True)

        cipd_package_subpath = "pigweed_internal/third_party/"
        cipd_package_subpath += core_name
        cipd_package_subpath += "/${platform}"

        # Check if teensy cipd package is readable

        with tempfile.NamedTemporaryFile(
            prefix='cipd', delete=True
        ) as temp_json:
            cipd_acl_check_command = [
                "cipd",
                "acl-check",
                cipd_package_subpath,
                "-reader",
                "-json-output",
                temp_json.name,
            ]
            subprocess.run(cipd_acl_check_command, capture_output=True)
            # Return if no packages are readable.
            if not json.load(temp_json)['result']:
                return

        def _run_command(command):
            _LOG.debug("Running: `%s`", " ".join(command))
            result = subprocess.run(command, capture_output=True)
            _LOG.debug(
                "Output:\n%s", result.stdout.decode() + result.stderr.decode()
            )

        _run_command(["cipd", "init", "-force", core_cache_path.as_posix()])
        _run_command(
            [
                "cipd",
                "install",
                cipd_package_subpath,
                "-root",
                core_cache_path.as_posix(),
                "-force",
            ]
        )

        _LOG.debug(
            "Available Cache Files:\n%s",
            "\n".join([p.as_posix() for p in core_cache_path.glob("*")]),
        )

    def install(self, path: Path) -> None:
        self.populate_download_cache_from_cipd(path)

        if self.status(path):
            return
        # Otherwise delete current version and reinstall
        core_installer.install_core(path.parent.resolve().as_posix(), self.name)

    def info(self, path: Path) -> Sequence[str]:
        packages_root = path.parent.resolve()
        arduino_package_path = path
        arduino_package_name = None

        message = [
            f'{self.name} currently installed in: {path}',
        ]
        # Make gn args sample copy/paste-able by omitting the starting timestamp
        # and INF log on each line.
        message_gn_args = [
            'Enable by running "gn args out" and adding these lines:',
            f'  pw_arduino_build_CORE_PATH = "{packages_root}"',
            f'  pw_arduino_build_CORE_NAME = "{self.name}"',
        ]

        # Search for first valid 'package/version' directory
        for hardware_dir in [
            path for path in (path / 'hardware').iterdir() if path.is_dir()
        ]:
            if path.name in ["arduino", "tools"]:
                continue
            for subdir in [
                path for path in hardware_dir.iterdir() if path.is_dir()
            ]:
                if subdir.name == 'avr' or re.match(r'[0-9.]+', subdir.name):
                    arduino_package_name = f'{hardware_dir.name}/{subdir.name}'
                    break

        if arduino_package_name:
            message_gn_args += [
                f'  pw_arduino_build_PACKAGE_NAME = "{arduino_package_name}"',
                '  pw_arduino_build_BOARD = "BOARD_NAME"',
            ]
            message += ["\n".join(message_gn_args)]
            message += [
                'Where BOARD_NAME is any supported board.',
                # Have arduino_builder command appear on it's own line.
                'List available boards by running:\n'
                '  arduino_builder '
                f'--arduino-package-path {arduino_package_path} '
                f'--arduino-package-name {arduino_package_name} list-boards',
            ]
        return message


for arduino_core_name in core_installer.supported_cores():
    pw_package.package_manager.register(
        ArduinoCore, core_name=arduino_core_name
    )



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""This module generates the code for nanopb-based pw_rpc services."""

import os
from typing import Iterable, NamedTuple, Optional

from pw_protobuf.output_file import OutputFile
from pw_protobuf.proto_tree import ProtoServiceMethod
from pw_protobuf.proto_tree import build_node_tree
from pw_rpc import codegen
from pw_rpc.codegen import (
    client_call_type,
    get_id,
    CodeGenerator,
    RPC_NAMESPACE,
)

PROTO_H_EXTENSION = '.pb.h'
PROTO_CC_EXTENSION = '.pb.cc'
NANOPB_H_EXTENSION = '.pb.h'


def _serde(method: ProtoServiceMethod) -> str:
    """Returns the NanopbMethodSerde for this method."""
    return (
        f'{RPC_NAMESPACE}::internal::kNanopbMethodSerde<'
        f'{method.request_type().nanopb_fields()}, '
        f'{method.response_type().nanopb_fields()}>'
    )


def _proto_filename_to_nanopb_header(proto_file: str) -> str:
    """Returns the generated nanopb header name for a .proto file."""
    return os.path.splitext(proto_file)[0] + NANOPB_H_EXTENSION


def _proto_filename_to_generated_header(proto_file: str) -> str:
    """Returns the generated C++ RPC header name for a .proto file."""
    filename = os.path.splitext(proto_file)[0]
    return f'{filename}.rpc{PROTO_H_EXTENSION}'


def _client_call(method: ProtoServiceMethod) -> str:
    template_args = []

    if method.client_streaming():
        template_args.append(method.request_type().nanopb_struct())

    template_args.append(method.response_type().nanopb_struct())

    return f'{client_call_type(method, "Nanopb")}<{", ".join(template_args)}>'


def _function(method: ProtoServiceMethod) -> str:
    return f'{_client_call(method)} {method.name()}'


def _user_args(method: ProtoServiceMethod) -> Iterable[str]:
    if not method.client_streaming():
        yield f'const {method.request_type().nanopb_struct()}& request'

    response = method.response_type().nanopb_struct()

    if method.server_streaming():
        yield f'::pw::Function<void(const {response}&)>&& on_next = nullptr'
        yield '::pw::Function<void(::pw::Status)>&& on_completed = nullptr'
    else:
        yield (
            f'::pw::Function<void(const {response}&, ::pw::Status)>&& '
            'on_completed = nullptr'
        )

    yield '::pw::Function<void(::pw::Status)>&& on_error = nullptr'


class NanopbCodeGenerator(CodeGenerator):
    """Generates an RPC service and client using the Nanopb API."""

    def name(self) -> str:
        return 'nanopb'

    def method_union_name(self) -> str:
        return 'NanopbMethodUnion'

    def includes(self, proto_file_name: str) -> Iterable[str]:
        yield '#include "pw_rpc/nanopb/client_reader_writer.h"'
        yield '#include "pw_rpc/nanopb/internal/method_union.h"'
        yield '#include "pw_rpc/nanopb/server_reader_writer.h"'

        # Include the corresponding nanopb header file for this proto file, in
        # which the file's messages and enums are generated. All other files
        # imported from the .proto file are #included in there.
        nanopb_header = _proto_filename_to_nanopb_header(proto_file_name)
        yield f'#include "{nanopb_header}"'

    def service_aliases(self) -> None:
        self.line('template <typename Response>')
        self.line(
            'using ServerWriter = '
            f'{RPC_NAMESPACE}::NanopbServerWriter<Response>;'
        )
        self.line('template <typename Request, typename Response>')
        self.line(
            'using ServerReader = '
            f'{RPC_NAMESPACE}::NanopbServerReader<Request, Response>;'
        )
        self.line('template <typename Request, typename Response>')
        self.line(
            'using ServerReaderWriter = '
            f'{RPC_NAMESPACE}::NanopbServerReaderWriter<Request, Response>;'
        )

    def method_descriptor(self, method: ProtoServiceMethod) -> None:
        self.line(
            f'{RPC_NAMESPACE}::internal::'
            f'GetNanopbOrRawMethodFor<&Implementation::{method.name()}, '
            f'{method.type().cc_enum()}, '
            f'{method.request_type().nanopb_struct()}, '
            f'{method.response_type().nanopb_struct()}>('
        )
        with self.indent(4):
            self.line(f'{get_id(method)},  // Hash of "{method.name()}"')
            self.line(f'{_serde(method)}),')

    def client_member_function(self, method: ProtoServiceMethod) -> None:
        """Outputs client code for a single RPC method."""

        self.line(f'{_function(method)}(')
        self.indented_list(*_user_args(method), end=') const {')

        with self.indent():
            client_call = _client_call(method)
            base = 'Stream' if method.server_streaming() else 'Unary'
            self.line(
                f'return {RPC_NAMESPACE}::internal::'
                f'Nanopb{base}ResponseClientCall<'
                f'{method.response_type().nanopb_struct()}>::'
                f'Start<{client_call}>('
            )

            service_client = RPC_NAMESPACE + '::internal::ServiceClient'

            args = [
                f'{service_client}::client()',
                f'{service_client}::channel_id()',
                'kServiceId',
                get_id(method),
                _serde(method),
            ]
            if method.server_streaming():
                args.append('std::move(on_next)')

            args.append('std::move(on_completed)')
            args.append('std::move(on_error)')

            if not method.client_streaming():
                args.append('request')

            self.indented_list(*args, end=');')

        self.line('}')

    def client_static_function(self, method: ProtoServiceMethod) -> None:
        self.line(f'static {_function(method)}(')
        self.indented_list(
            f'{RPC_NAMESPACE}::Client& client',
            'uint32_t channel_id',
            *_user_args(method),
            end=') {',
        )

        with self.indent():
            self.line(f'return Client(client, channel_id).{method.name()}(')

            args = []

            if not method.client_streaming():
                args.append('request')

            if method.server_streaming():
                args.append('std::move(on_next)')

            self.indented_list(
                *args,
                'std::move(on_completed)',
                'std::move(on_error)',
                end=');',
            )

        self.line('}')

    def method_info_specialization(self, method: ProtoServiceMethod) -> None:
        self.line()
        self.line(f'using Request = {method.request_type().nanopb_struct()};')
        self.line(f'using Response = {method.response_type().nanopb_struct()};')
        self.line()
        self.line(
            f'static constexpr const {RPC_NAMESPACE}::internal::'
            'NanopbMethodSerde& serde() {'
        )
        with self.indent():
            self.line(f'return {_serde(method)};')
        self.line('}')


class _CallbackFunction(NamedTuple):
    """Represents a callback function parameter in a client RPC call."""

    function_type: str
    name: str
    default_value: Optional[str] = None

    def __str__(self):
        param = f'::pw::Function<{self.function_type}>&& {self.name}'
        if self.default_value:
            param += f' = {self.default_value}'
        return param


class StubGenerator(codegen.StubGenerator):
    """Generates Nanopb RPC stubs."""

    def unary_signature(self, method: ProtoServiceMethod, prefix: str) -> str:
        return (
            f'::pw::Status {prefix}{method.name()}( '
            f'const {method.request_type().nanopb_struct()}& request, '
            f'{method.response_type().nanopb_struct()}& response)'
        )

    def unary_stub(
        self, method: ProtoServiceMethod, output: OutputFile
    ) -> None:
        output.write_line(codegen.STUB_REQUEST_TODO)
        output.write_line('static_cast<void>(request);')
        output.write_line(codegen.STUB_RESPONSE_TODO)
        output.write_line('static_cast<void>(response);')
        output.write_line('return ::pw::Status::Unimplemented();')

    def server_streaming_signature(
        self, method: ProtoServiceMethod, prefix: str
    ) -> str:
        return (
            f'void {prefix}{method.name()}( '
            f'const {method.request_type().nanopb_struct()}& request, '
            f'ServerWriter<{method.response_type().nanopb_struct()}>& writer)'
        )

    def client_streaming_signature(
        self, method: ProtoServiceMethod, prefix: str
    ) -> str:
        return (
            f'void {prefix}{method.name()}( '
            f'ServerReader<{method.request_type().nanopb_struct()}, '
            f'{method.response_type().nanopb_struct()}>& reader)'
        )

    def bidirectional_streaming_signature(
        self, method: ProtoServiceMethod, prefix: str
    ) -> str:
        return (
            f'void {prefix}{method.name()}( '
            f'ServerReaderWriter<{method.request_type().nanopb_struct()}, '
            f'{method.response_type().nanopb_struct()}>& reader_writer)'
        )


def process_proto_file(proto_file) -> Iterable[OutputFile]:
    """Generates code for a single .proto file."""

    _, package_root = build_node_tree(proto_file)
    output_filename = _proto_filename_to_generated_header(proto_file.name)
    generator = NanopbCodeGenerator(output_filename)
    codegen.generate_package(proto_file, package_root, generator)

    codegen.package_stubs(package_root, generator, StubGenerator())

    return [generator.output]



#!/usr/bin/env python3
# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Arduino Core Installer."""

import argparse
import logging
import operator
import os
import platform
import shutil
import stat
import subprocess
import sys
import time
from pathlib import Path
from typing import Dict, List

from pw_arduino_build import file_operations

_LOG = logging.getLogger(__name__)


class ArduinoCoreNotSupported(Exception):
    """Exception raised when a given core can not be installed."""


_ARDUINO_CORE_ARTIFACTS: Dict[str, Dict] = {
    # pylint: disable=line-too-long
    "teensy": {
        "Linux": {
            "arduino-ide": {
                "url": "https://downloads.arduino.cc/arduino-1.8.13-linux64.tar.xz",
                "file_name": "arduino-1.8.13-linux64.tar.xz",
                "sha256": "1b20d0ec850a2a63488009518725f058668bb6cb48c321f82dcf47dc4299b4ad",
            },
            "teensyduino": {
                "url": "https://www.pjrc.com/teensy/td_153/TeensyduinoInstall.linux64",
                "sha256": "2e6cd99a757bc80593ea3de006de4cc934bcb0a6ec74cad8ec327f0289d40f0b",
                "file_name": "TeensyduinoInstall.linux64",
            },
        },
        # TODO(tonymd): Handle 32-bit Linux Install?
        "Linux32": {
            "arduino-ide": {
                "url": "https://downloads.arduino.cc/arduino-1.8.13-linux32.tar.xz",
                "file_name": "arduino-1.8.13-linux32.tar.xz",
                "sha256": "",
            },
            "teensyduino": {
                "url": "https://www.pjrc.com/teensy/td_153/TeensyduinoInstall.linux32",
                "file_name": "TeensyduinoInstall.linux32",
                "sha256": "",
            },
        },
        # TODO(tonymd): Handle ARM32 (Raspberry Pi) Install?
        "LinuxARM32": {
            "arduino-ide": {
                "url": "https://downloads.arduino.cc/arduino-1.8.13-linuxarm.tar.xz",
                "file_name": "arduino-1.8.13-linuxarm.tar.xz",
                "sha256": "",
            },
            "teensyduino": {
                "url": "https://www.pjrc.com/teensy/td_153/TeensyduinoInstall.linuxarm",
                "file_name": "TeensyduinoInstall.linuxarm",
                "sha256": "",
            },
        },
        # TODO(tonymd): Handle ARM64 Install?
        "LinuxARM64": {
            "arduino-ide": {
                "url": "https://downloads.arduino.cc/arduino-1.8.13-linuxaarch64.tar.xz",
                "file_name": "arduino-1.8.13-linuxaarch64.tar.xz",
                "sha256": "",
            },
            "teensyduino": {
                "url": "https://www.pjrc.com/teensy/td_153/TeensyduinoInstall.linuxaarch64",
                "file_name": "TeensyduinoInstall.linuxaarch64",
                "sha256": "",
            },
        },
        "Darwin": {
            "teensyduino": {
                "url": "https://www.pjrc.com/teensy/td_153/Teensyduino_MacOS_Catalina.zip",
                "file_name": "Teensyduino_MacOS_Catalina.zip",
                "sha256": "401ef42c6e83e621cdda20191a4ef9b7db8a214bede5a94a9e26b45f79c64fe2",
            },
        },
        "Windows": {
            "arduino-ide": {
                "url": "https://downloads.arduino.cc/arduino-1.8.13-windows.zip",
                "file_name": "arduino-1.8.13-windows.zip",
                "sha256": "78d3e96827b9e9b31b43e516e601c38d670d29f12483e88cbf6d91a0f89ef524",
            },
            "teensyduino": {
                "url": "https://www.pjrc.com/teensy/td_153/TeensyduinoInstall.exe",
                # The installer should be named 'Teensyduino.exe' instead of
                # 'TeensyduinoInstall.exe' to trigger a non-admin installation.
                "file_name": "Teensyduino.exe",
                "sha256": "88f58681e5c4772c54e462bc88280320e4276e5b316dcab592fe38d96db990a1",
            },
        },
    },
    "adafruit-samd": {
        "all": {
            "core": {
                "version": "1.6.2",
                "url": "https://github.com/adafruit/ArduinoCore-samd/archive/1.6.2.tar.gz",
                "file_name": "adafruit-samd-1.6.2.tar.gz",
                "sha256": "5875f5bc05904c10e6313f02653f28f2f716db639d3d43f5a1d8a83d15339d64",
            }
        },
        "Linux": {},
        "Darwin": {},
        "Windows": {},
    },
    "arduino-samd": {
        "all": {
            "core": {
                "version": "1.8.8",
                "url": "http://downloads.arduino.cc/cores/samd-1.8.8.tar.bz2",
                "file_name": "arduino-samd-1.8.8.tar.bz2",
                "sha256": "7b93eb705cba9125d9ee52eba09b51fb5fe34520ada351508f4253abbc9f27fa",
            }
        },
        "Linux": {},
        "Darwin": {},
        "Windows": {},
    },
    "stm32duino": {
        "all": {
            "core": {
                "version": "1.9.0",
                "url": "https://github.com/stm32duino/Arduino_Core_STM32/archive/1.9.0.tar.gz",
                "file_name": "stm32duino-1.9.0.tar.gz",
                "sha256": "4f75ba7a117d90392e8f67c58d31d22393749b9cdd3279bc21e7261ec06c62bf",
            }
        },
        "Linux": {},
        "Darwin": {},
        "Windows": {},
    },
}


def install_core_command(args: argparse.Namespace):
    install_core(args.prefix, args.core_name)


def install_core(prefix, core_name):
    install_prefix = os.path.realpath(
        os.path.expanduser(os.path.expandvars(prefix))
    )
    install_dir = os.path.join(install_prefix, core_name)
    cache_dir = os.path.join(install_prefix, ".cache", core_name)

    if core_name in supported_cores():
        shutil.rmtree(install_dir, ignore_errors=True)
        os.makedirs(install_dir, exist_ok=True)
        os.makedirs(cache_dir, exist_ok=True)

    if core_name == "teensy":
        if platform.system() == "Linux":
            install_teensy_core_linux(install_prefix, install_dir, cache_dir)
        elif platform.system() == "Darwin":
            install_teensy_core_mac(install_prefix, install_dir, cache_dir)
        elif platform.system() == "Windows":
            install_teensy_core_windows(install_prefix, install_dir, cache_dir)
        apply_teensy_patches(install_dir)
    elif core_name == "adafruit-samd":
        install_adafruit_samd_core(install_prefix, install_dir, cache_dir)
    elif core_name == "stm32duino":
        install_stm32duino_core(install_prefix, install_dir, cache_dir)
    elif core_name == "arduino-samd":
        install_arduino_samd_core(install_prefix, install_dir, cache_dir)
    else:
        raise ArduinoCoreNotSupported(
            "Invalid core '{}'. Supported cores: {}".format(
                core_name, ", ".join(supported_cores())
            )
        )


def supported_cores():
    return _ARDUINO_CORE_ARTIFACTS.keys()


def get_windows_process_names() -> List[str]:
    result = subprocess.run("wmic process get description", capture_output=True)
    output = result.stdout.decode().splitlines()
    return [line.strip() for line in output if line]


def install_teensy_core_windows(install_prefix, install_dir, cache_dir):
    """Download and install Teensyduino artifacts for Windows."""
    teensy_artifacts = _ARDUINO_CORE_ARTIFACTS["teensy"][platform.system()]

    arduino_artifact = teensy_artifacts["arduino-ide"]
    arduino_zipfile = file_operations.download_to_cache(
        url=arduino_artifact["url"],
        expected_sha256sum=arduino_artifact["sha256"],
        cache_directory=cache_dir,
        downloaded_file_name=arduino_artifact["file_name"],
    )

    teensyduino_artifact = teensy_artifacts["teensyduino"]
    teensyduino_installer = file_operations.download_to_cache(
        url=teensyduino_artifact["url"],
        expected_sha256sum=teensyduino_artifact["sha256"],
        cache_directory=cache_dir,
        downloaded_file_name=teensyduino_artifact["file_name"],
    )

    file_operations.extract_archive(arduino_zipfile, install_dir, cache_dir)

    # "teensy" here should match args.core_name
    teensy_core_dir = os.path.join(install_prefix, "teensy")

    # Change working directory for installation
    original_working_dir = os.getcwd()
    os.chdir(install_prefix)

    install_command = [teensyduino_installer, "--dir=teensy"]
    _LOG.info("  Running: %s", " ".join(install_command))
    _LOG.info(
        "    Please click yes on the Windows 'User Account Control' " "dialog."
    )
    _LOG.info("    You should see: 'Verified publisher: PRJC.COM LLC'")

    def wait_for_process(
        process_name, timeout=30, result_operator=operator.truth
    ):
        start_time = time.time()
        while result_operator(process_name in get_windows_process_names()):
            time.sleep(1)
            if time.time() > start_time + timeout:
                _LOG.error(
                    "Error: Installation Failed.\n"
                    "Please click yes on the Windows 'User Account Control' "
                    "dialog."
                )
                sys.exit(1)

    # Run Teensyduino installer with admin rights (non-blocking)
    # User Account Control (UAC) will prompt the user for consent
    import ctypes  # pylint: disable=import-outside-toplevel

    ctypes.windll.shell32.ShellExecuteW(
        None,  # parent window handle
        "runas",  # operation
        teensyduino_installer,  # file to run
        subprocess.list2cmdline(install_command),  # command parameters
        install_prefix,  # working directory
        1,
    )  # Display mode (SW_SHOWNORMAL: Activates and displays a window)

    # Wait for teensyduino_installer to start running
    wait_for_process("TeensyduinoInstall.exe", result_operator=operator.not_)

    _LOG.info("Waiting for TeensyduinoInstall.exe to finish.")
    # Wait till teensyduino_installer is finished
    wait_for_process("TeensyduinoInstall.exe", timeout=360)

    if not os.path.exists(os.path.join(teensy_core_dir, "hardware", "teensy")):
        _LOG.error(
            "Error: Installation Failed.\n"
            "Please try again and ensure Teensyduino is installed in "
            "the folder:\n"
            "%s",
            teensy_core_dir,
        )
        sys.exit(1)
    else:
        _LOG.info("Install complete!")

    file_operations.remove_empty_directories(install_dir)
    os.chdir(original_working_dir)


def install_teensy_core_mac(unused_install_prefix, install_dir, cache_dir):
    """Download and install Teensyduino artifacts for Mac."""
    teensy_artifacts = _ARDUINO_CORE_ARTIFACTS["teensy"][platform.system()]

    teensyduino_artifact = teensy_artifacts["teensyduino"]
    teensyduino_zip = file_operations.download_to_cache(
        url=teensyduino_artifact["url"],
        expected_sha256sum=teensyduino_artifact["sha256"],
        cache_directory=cache_dir,
        downloaded_file_name=teensyduino_artifact["file_name"],
    )

    extracted_files = file_operations.extract_archive(
        teensyduino_zip,
        install_dir,
        cache_dir,
        remove_single_toplevel_folder=False,
    )
    toplevel_folder = sorted(extracted_files)[0]
    os.symlink(
        os.path.join(toplevel_folder, "Contents", "Java", "hardware"),
        os.path.join(install_dir, "hardware"),
        target_is_directory=True,
    )


def install_teensy_core_linux(install_prefix, install_dir, cache_dir):
    """Download and install Teensyduino artifacts for Windows."""
    teensy_artifacts = _ARDUINO_CORE_ARTIFACTS["teensy"][platform.system()]

    arduino_artifact = teensy_artifacts["arduino-ide"]
    arduino_tarfile = file_operations.download_to_cache(
        url=arduino_artifact["url"],
        expected_sha256sum=arduino_artifact["sha256"],
        cache_directory=cache_dir,
        downloaded_file_name=arduino_artifact["file_name"],
    )

    teensyduino_artifact = teensy_artifacts["teensyduino"]
    teensyduino_installer = file_operations.download_to_cache(
        url=teensyduino_artifact["url"],
        expected_sha256sum=teensyduino_artifact["sha256"],
        cache_directory=cache_dir,
        downloaded_file_name=teensyduino_artifact["file_name"],
    )

    file_operations.extract_archive(arduino_tarfile, install_dir, cache_dir)
    os.chmod(
        teensyduino_installer,
        os.stat(teensyduino_installer).st_mode | stat.S_IEXEC,
    )

    original_working_dir = os.getcwd()
    os.chdir(install_prefix)
    # "teensy" here should match args.core_name
    install_command = [teensyduino_installer, "--dir=teensy"]
    subprocess.run(install_command)

    file_operations.remove_empty_directories(install_dir)
    os.chdir(original_working_dir)


def apply_teensy_patches(install_dir):
    # On Mac the "hardware" directory is a symlink:
    #   ls -l third_party/arduino/cores/teensy/
    #   hardware -> Teensyduino.app/Contents/Java/hardware
    # Resolve paths since `git apply` doesn't work if a path is beyond a
    # symbolic link.
    patch_root_path = (
        Path(install_dir) / "hardware/teensy/avr/cores"
    ).resolve()

    # Get all *.diff files relative to this python file's parent directory.
    patch_file_paths = sorted(
        (Path(__file__).parent / "core_patches/teensy").glob("*.diff")
    )

    # Apply each patch file.
    for diff_path in patch_file_paths:
        file_operations.git_apply_patch(
            patch_root_path.as_posix(), diff_path.as_posix(), unsafe_paths=True
        )


def install_arduino_samd_core(
    install_prefix: str, install_dir: str, cache_dir: str
):
    artifacts = _ARDUINO_CORE_ARTIFACTS["arduino-samd"]["all"]["core"]
    core_tarfile = file_operations.download_to_cache(
        url=artifacts["url"],
        expected_sha256sum=artifacts["sha256"],
        cache_directory=cache_dir,
        downloaded_file_name=artifacts["file_name"],
    )

    package_path = os.path.join(
        install_dir, "hardware", "samd", artifacts["version"]
    )
    os.makedirs(package_path, exist_ok=True)
    file_operations.extract_archive(core_tarfile, package_path, cache_dir)
    original_working_dir = os.getcwd()
    os.chdir(install_prefix)
    # TODO(tonymd): Fetch core/tools as specified by:
    # http://downloads.arduino.cc/packages/package_index.json
    os.chdir(original_working_dir)
    return True


def install_adafruit_samd_core(
    install_prefix: str, install_dir: str, cache_dir: str
):
    artifacts = _ARDUINO_CORE_ARTIFACTS["adafruit-samd"]["all"]["core"]
    core_tarfile = file_operations.download_to_cache(
        url=artifacts["url"],
        expected_sha256sum=artifacts["sha256"],
        cache_directory=cache_dir,
        downloaded_file_name=artifacts["file_name"],
    )

    package_path = os.path.join(
        install_dir, "hardware", "samd", artifacts["version"]
    )
    os.makedirs(package_path, exist_ok=True)
    file_operations.extract_archive(core_tarfile, package_path, cache_dir)

    original_working_dir = os.getcwd()
    os.chdir(install_prefix)
    # TODO(tonymd): Fetch platform specific tools as specified by:
    # https://adafruit.github.io/arduino-board-index/package_adafruit_index.json
    # Specifically:
    #   https://github.com/ARM-software/CMSIS_5/archive/5.4.0.tar.gz
    os.chdir(original_working_dir)
    return True


def install_stm32duino_core(install_prefix, install_dir, cache_dir):
    artifacts = _ARDUINO_CORE_ARTIFACTS["stm32duino"]["all"]["core"]
    core_tarfile = file_operations.download_to_cache(
        url=artifacts["url"],
        expected_sha256sum=artifacts["sha256"],
        cache_directory=cache_dir,
        downloaded_file_name=artifacts["file_name"],
    )

    package_path = os.path.join(
        install_dir, "hardware", "stm32", artifacts["version"]
    )
    os.makedirs(package_path, exist_ok=True)
    file_operations.extract_archive(core_tarfile, package_path, cache_dir)
    original_working_dir = os.getcwd()
    os.chdir(install_prefix)
    # TODO(tonymd): Fetch platform specific tools as specified by:
    # https://github.com/stm32duino/BoardManagerFiles/raw/HEAD/STM32/package_stm_index.json
    os.chdir(original_working_dir)
    return True



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Generates compile fail test GN targets.

Scans source files for PW_NC_TEST(...) statements and generates a
BUILD.gn file with a target for each test. This allows the compilation failure
tests to run in parallel in Ninja.

This file is executed during gn gen, so it cannot rely on any setup that occurs
during the build.
"""

import argparse
import base64
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
import pickle
import re
import sys
from typing import (
    Iterable,
    Iterator,
    List,
    NamedTuple,
    NoReturn,
    Optional,
    Pattern,
    Sequence,
    Set,
    Tuple,
)

# Matches the #if or #elif statement that starts a compile fail test.
_TEST_START = re.compile(r'^[ \t]*#[ \t]*(?:el)?if[ \t]+PW_NC_TEST\([ \t]*')

# Matches the name of a test case.
_TEST_NAME = re.compile(
    r'(?P<name>[a-zA-Z0-9_]+)[ \t]*\)[ \t]*(?://.*|/\*.*)?$'
)

# Negative compilation test commands take the form PW_NC_EXPECT("regex"),
# PW_NC_EXPECT_GCC("regex"), or PW_NC_EXPECT_CLANG("regex"). PW_NC_EXPECT() is
# an error.
_EXPECT_START = re.compile(r'^[ \t]*PW_NC_EXPECT(?P<compiler>_GCC|_CLANG)?\(')

# EXPECT statements are regular expressions that must match the compiler output.
# They must fit on a single line.
_EXPECT_REGEX = re.compile(r'(?P<regex>"[^\n]+")\);[ \t]*(?://.*|/\*.*)?$')


class Compiler(Enum):
    ANY = 0
    GCC = 1
    CLANG = 2

    @staticmethod
    def from_command(command: str) -> 'Compiler':
        if command.endswith(('clang', 'clang++')):
            return Compiler.CLANG

        if command.endswith(('gcc', 'g++')):
            return Compiler.GCC

        raise ValueError(
            f"Unrecognized compiler '{command}'; update the Compiler enum "
            f'in {Path(__file__).name} to account for this'
        )

    def matches(self, other: 'Compiler') -> bool:
        return self is other or self is Compiler.ANY or other is Compiler.ANY


@dataclass(frozen=True)
class Expectation:
    compiler: Compiler
    pattern: Pattern[str]
    line: int


@dataclass(frozen=True)
class TestCase:
    suite: str
    case: str
    expectations: Tuple[Expectation, ...]
    source: Path
    line: int

    def name(self) -> str:
        return f'{self.suite}.{self.case}'

    def serialize(self) -> str:
        return base64.b64encode(pickle.dumps(self)).decode()

    @classmethod
    def deserialize(cls, serialized: str) -> 'Expectation':
        return pickle.loads(base64.b64decode(serialized))


class ParseError(Exception):
    """Failed to parse a PW_NC_TEST."""

    def __init__(
        self,
        message: str,
        file: Path,
        lines: Sequence[str],
        error_lines: Sequence[int],
    ) -> None:
        for i in error_lines:
            message += f'\n{file.name}:{i + 1}: {lines[i]}'
        super().__init__(message)


class _ExpectationParser:
    """Parses expecatations from 'PW_NC_EXPECT(' to the final ');'."""

    class _State:
        SPACE = 0  # Space characters, which are ignored
        COMMENT_START = 1  # First / in a //-style comment
        COMMENT = 2  # Everything after // on a line
        OPEN_QUOTE = 3  # Starting quote for a string literal
        CHARACTERS = 4  # Characters within a string literal
        ESCAPE = 5  # \ within a string literal, which may escape a "
        CLOSE_PAREN = 6  # Closing parenthesis to the PW_NC_EXPECT statement.

    def __init__(self, index: int, compiler: Compiler) -> None:
        self.index = index
        self._compiler = compiler
        self._state = self._State.SPACE
        self._contents: List[str] = []

    def parse(self, chars: str) -> Optional[Expectation]:
        """State machine that parses characters in PW_NC_EXPECT()."""
        for char in chars:
            if self._state is self._State.SPACE:
                if char == '"':
                    self._state = self._State.CHARACTERS
                elif char == ')':
                    self._state = self._State.CLOSE_PAREN
                elif char == '/':
                    self._state = self._State.COMMENT_START
                elif not char.isspace():
                    raise ValueError(f'Unexpected character "{char}"')
            elif self._state is self._State.COMMENT_START:
                if char == '*':
                    raise ValueError(
                        '"/* */" comments are not supported; use // instead'
                    )
                if char != '/':
                    raise ValueError(f'Unexpected character "{char}"')
                self._state = self._State.COMMENT
            elif self._state is self._State.COMMENT:
                if char == '\n':
                    self._state = self._State.SPACE
            elif self._state is self._State.CHARACTERS:
                if char == '"':
                    self._state = self._State.SPACE
                elif char == '\\':
                    self._state = self._State.ESCAPE
                else:
                    self._contents.append(char)
            elif self._state is self._State.ESCAPE:
                # Include escaped " directly. Restore the \ for other chars.
                if char != '"':
                    self._contents.append('\\')
                self._contents.append(char)
                self._state = self._State.CHARACTERS
            elif self._state is self._State.CLOSE_PAREN:
                if char != ';':
                    raise ValueError(f'Expected ";", found "{char}"')

                return self._expectation(''.join(self._contents))

        return None

    def _expectation(self, regex: str) -> Expectation:
        if '"""' in regex:
            raise ValueError('The regular expression cannot contain """')

        # Evaluate the string from the C++ source as a raw literal.
        re_string = eval(f'r"""{regex}"""')  # pylint: disable=eval-used
        if not isinstance(re_string, str):
            raise ValueError('The regular expression must be a string!')

        try:
            return Expectation(
                self._compiler, re.compile(re_string), self.index + 1
            )
        except re.error as error:
            raise ValueError('Invalid regular expression: ' + error.msg)


class _NegativeCompilationTestSource:
    def __init__(self, file: Path) -> None:
        self._file = file
        self._lines = self._file.read_text().splitlines(keepends=True)

        self._parsed_expectations: Set[int] = set()

    def _error(self, message: str, *error_lines: int) -> NoReturn:
        raise ParseError(message, self._file, self._lines, error_lines)

    def _parse_expectations(self, start: int) -> Iterator[Expectation]:
        expectation: Optional[_ExpectationParser] = None

        for index in range(start, len(self._lines)):
            line = self._lines[index]

            # Skip empty or comment lines
            if not line or line.isspace() or line.lstrip().startswith('//'):
                continue

            # Look for a 'PW_NC_EXPECT(' in the code.
            if not expectation:
                expect_match = _EXPECT_START.match(line)
                if not expect_match:
                    break  # No expectation found, stop processing.

                compiler = expect_match['compiler'] or 'ANY'
                expectation = _ExpectationParser(
                    index, Compiler[compiler.lstrip('_')]
                )

                self._parsed_expectations.add(index)

                # Remove the 'PW_NC_EXPECT(' so the line starts with the regex.
                line = line[expect_match.end() :]

            # Find the regex after previously finding 'PW_NC_EXPECT('.
            try:
                if parsed_expectation := expectation.parse(line.lstrip()):
                    yield parsed_expectation

                    expectation = None
            except ValueError as err:
                self._error(
                    f'Failed to parse PW_NC_EXPECT() statement:\n\n  {err}.\n\n'
                    'PW_NC_EXPECT() statements must contain only a string '
                    'literal with a valid Python regular expression and '
                    'optional //-style comments.',
                    index,
                )

        if expectation:
            self._error(
                'Unterminated PW_NC_EXPECT() statement!', expectation.index
            )

    def _check_for_stray_expectations(self) -> None:
        all_expectations = frozenset(
            i
            for i in range(len(self._lines))
            if _EXPECT_START.match(self._lines[i])
        )
        stray = all_expectations - self._parsed_expectations
        if stray:
            self._error(
                f'Found {len(stray)} stray PW_NC_EXPECT() commands!',
                *sorted(stray),
            )

    def parse(self, suite: str) -> Iterator[TestCase]:
        """Finds all negative compilation tests in this source file."""
        for index, line in enumerate(self._lines):
            case_match = _TEST_START.match(line)
            if not case_match:
                continue

            name_match = _TEST_NAME.match(line, case_match.end())
            if not name_match:
                self._error(
                    'Negative compilation test syntax error. '
                    f"Expected test name, found '{line[case_match.end():]}'",
                    index,
                )

            expectations = tuple(self._parse_expectations(index + 1))
            yield TestCase(
                suite, name_match['name'], expectations, self._file, index + 1
            )

        self._check_for_stray_expectations()


def enumerate_tests(suite: str, paths: Iterable[Path]) -> Iterator[TestCase]:
    """Parses PW_NC_TEST statements from a file."""
    for path in paths:
        yield from _NegativeCompilationTestSource(path).parse(suite)


class SourceFile(NamedTuple):
    gn_path: str
    file_path: Path


def generate_gn_target(
    base: str, source_list: str, test: TestCase, all_tests: str
) -> Iterator[str]:
    yield f'''\
pw_python_action("{test.name()}.negative_compilation_test") {{
  script = "$dir_pw_compilation_testing/py/pw_compilation_testing/runner.py"
  inputs = [{source_list}]
  args = [
    "--toolchain-ninja=$_toolchain_ninja",
    "--target-ninja=$_target_ninja",
    "--test-data={test.serialize()}",
    "--all-tests={all_tests}",
  ]
  deps = ["{base}"]
  python_deps = [
    "$dir_pw_cli/py",
    "$dir_pw_compilation_testing/py",
  ]
  stamp = true
}}
'''


def generate_gn_build(
    base: str,
    sources: Iterable[SourceFile],
    tests: List[TestCase],
    all_tests: str,
) -> Iterator[str]:
    """Generates the BUILD.gn file with compilation failure test targets."""
    _, base_name = base.rsplit(':', 1)

    yield 'import("//build_overrides/pigweed.gni")'
    yield ''
    yield 'import("$dir_pw_build/python_action.gni")'
    yield ''
    yield (
        '_toolchain_ninja = '
        'rebase_path("$root_out_dir/toolchain.ninja", root_build_dir)'
    )
    yield (
        '_target_ninja = '
        f'rebase_path(get_label_info("{base}", "target_out_dir") +'
        f'"/{base_name}.ninja", root_build_dir)'
    )
    yield ''

    gn_source_list = ', '.join(f'"{gn_path}"' for gn_path, _ in sources)
    for test in tests:
        yield from generate_gn_target(base, gn_source_list, test, all_tests)


def _main(
    name: str, base: str, sources: Iterable[SourceFile], output: Path
) -> int:
    print_stderr = lambda s: print(s, file=sys.stderr)

    try:
        tests = list(enumerate_tests(name, (s.file_path for s in sources)))
    except ParseError as error:
        print_stderr(f'ERROR: {error}')
        return 1

    if not tests:
        print_stderr(f'The test "{name}" has no negative compilation tests!')
        print_stderr(
            'Add PW_NC_TEST() cases or remove this negative ' 'compilation test'
        )
        return 1

    tests_by_case = defaultdict(list)
    for test in tests:
        tests_by_case[test.case].append(test)

    duplicates = [tests for tests in tests_by_case.values() if len(tests) > 1]
    if duplicates:
        print_stderr('There are duplicate negative compilation test cases!')
        print_stderr('The following test cases appear more than once:')
        for tests in duplicates:
            print_stderr(f'\n    {tests[0].case} ({len(tests)} occurrences):')
            for test in tests:
                print_stderr(f'        {test.source.name}:{test.line}')
        return 1

    output.mkdir(parents=True, exist_ok=True)
    build_gn = output.joinpath('BUILD.gn')
    with build_gn.open('w') as fd:
        for line in generate_gn_build(
            base, sources, tests, output.joinpath('tests.txt').as_posix()
        ):
            print(line, file=fd)

    with output.joinpath('tests.txt').open('w') as fd:
        for test in tests:
            print(test.case, file=fd)

    # Print the test case names to stdout for consumption by GN.
    for test in tests:
        print(test.case)

    return 0


def _parse_args() -> dict:
    """Parses command-line arguments."""

    def source_file(arg: str) -> SourceFile:
        gn_path, file_path = arg.split(';', 1)
        return SourceFile(gn_path, Path(file_path))

    parser = argparse.ArgumentParser(
        description='Emits an error when a facade has a null backend'
    )
    parser.add_argument('--output', type=Path, help='Output directory')
    parser.add_argument('--name', help='Name of the NC test')
    parser.add_argument('--base', help='GN label for the base target to build')
    parser.add_argument(
        'sources',
        nargs='+',
        type=source_file,
        help='Source file with the no-compile tests',
    )
    return vars(parser.parse_args())


if __name__ == '__main__':
    sys.exit(_main(**_parse_args()))



#!/usr/bin/env python3
# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Launch a pw_target_runner client that sends a test request."""

import argparse
import subprocess
import sys
from typing import Optional

_TARGET_CLIENT_COMMAND = 'pw_target_runner_client'


def parse_args():
    """Parses command-line arguments."""

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument('binary', help='The target test binary to run')
    parser.add_argument(
        '--server-port',
        type=int,
        default=8081,
        help='Port the test server is located on',
    )
    parser.add_argument(
        'runner_args',
        nargs=argparse.REMAINDER,
        help='Arguments to forward to the test runner',
    )

    return parser.parse_args()


def launch_client(binary: str, server_port: Optional[int]) -> int:
    """Sends a test request to the specified server port."""
    cmd = [_TARGET_CLIENT_COMMAND, '-binary', binary]

    if server_port is not None:
        cmd.extend(['-port', str(server_port)])

    return subprocess.call(cmd)


def main() -> int:
    """Launch a test by sending a request to a pw_target_runner_server."""
    args = parse_args()
    return launch_client(args.binary, args.server_port)


if __name__ == '__main__':
    sys.exit(main())



# Copyright 2019 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Prints an error for an incorrect usage of a pw_proto_library template."""

import argparse
import logging
import sys

from typing import Optional

_LOG = logging.getLogger(__name__)


def argument_parser(
    parser: Optional[argparse.ArgumentParser] = None,
) -> argparse.ArgumentParser:
    """Registers the script's arguments on an argument parser."""

    if parser is None:
        parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument('--dir', required=True, help='Target directory')
    parser.add_argument('--root', required=True, help='GN root')
    parser.add_argument('--target', required=True, help='Build target')
    parser.add_argument(
        'generators',
        metavar='GEN',
        nargs='+',
        help='Supported protobuf generators',
    )

    return parser


def main() -> int:
    """Prints an error message."""

    args = argument_parser().parse_args()
    relative_dir = args.dir[len(args.root) :].rstrip('/')

    _LOG.error('')
    _LOG.error('The target %s is not a compiled protobuf library.', args.target)
    _LOG.error('')
    _LOG.error('A different target is generated for each active generator.')
    _LOG.error('Depend on one of the following targets instead:')
    _LOG.error('')
    for gen in args.generators:
        _LOG.error('  //%s:%s.%s', relative_dir, args.target, gen)
    _LOG.error('')

    return 1


if __name__ == '__main__':
    try:
        # If pw_cli is available, use it to initialize logs.
        from pw_cli import log

        log.install(logging.INFO)
    except ImportError:
        # If pw_cli isn't available, display log messages like a simple print.
        logging.basicConfig(format='%(message)s', level=logging.INFO)

    sys.exit(main())



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""CLI tools for pw_ide."""

import argparse
import enum
from inspect import cleandoc
from pathlib import Path
import re
from typing import Any, Callable, Dict, List, Optional, Protocol

from pw_ide.commands import (
    cmd_clear,
    cmd_cpp,
    cmd_python,
    cmd_reset,
    cmd_setup,
    cmd_vscode,
)

from pw_ide.vscode import VscSettingsType


def _get_docstring(obj: Any) -> Optional[str]:
    doc: Optional[str] = getattr(obj, '__doc__', None)
    return doc


class _ParsedDocstring:
    """Parses help content out of a standard docstring."""

    def __init__(self, obj: Any) -> None:
        self.description = ''
        self.epilog = ''

        if obj is not None and (doc := _get_docstring(obj)) is not None:
            lines = doc.split('\n')
            self.description = lines.pop(0)

            # Eliminate the blank line between the summary and the main content
            if len(lines) > 0:
                lines.pop(0)

            self.epilog = cleandoc('\n'.join(lines))


class SphinxStripperState(enum.Enum):
    SEARCHING = 0
    COLLECTING = 1
    HANDLING = 2


class SphinxStripper:
    """Strip Sphinx directives from text.

    The caller can provide an object with methods named _handle_directive_{}
    to handle specific directives. Otherwise, the default will apply.

    Feed text line by line to .process(line), then get the processed text back
    with .result().
    """

    def __init__(self, handler: Any) -> None:
        self.handler = handler
        self.directive: str = ''
        self.tag: str = ''
        self.lines_to_handle: List[str] = []
        self.handled_lines: List[str] = []
        self._prev_state: SphinxStripperState = SphinxStripperState.SEARCHING
        self._curr_state: SphinxStripperState = SphinxStripperState.SEARCHING

    @property
    def state(self) -> SphinxStripperState:
        return self._curr_state

    @state.setter
    def state(self, value: SphinxStripperState) -> None:
        self._prev_state = self._curr_state
        self._curr_state = value

    def search_for_directives(self, line: str) -> None:
        match = re.search(
            r'^\.\.\s*(?P<directive>[\-\w]+)::\s*(?P<tag>[\-\w]+)$', line
        )

        if match is not None:
            self.directive = match.group('directive')
            self.tag = match.group('tag')
            self.state = SphinxStripperState.COLLECTING
        else:
            self.handled_lines.append(line)

    def collect_lines(self, line) -> None:
        # Collect lines associated with a directive, including blank lines in
        # the middle of the directive text, but not the blank line between the
        # directive and the start of its text.
        if not (line.strip() == '' and len(self.lines_to_handle) == 0):
            self.lines_to_handle.append(line)

    def handle_lines(self, line: str = '') -> None:
        handler_fn = f'_handle_directive_{self.directive.replace("-", "_")}'

        self.handled_lines.extend(
            getattr(self.handler, handler_fn, lambda _, s: s)(
                self.tag, self.lines_to_handle
            )
        )

        self.handled_lines.append(line)
        self.lines_to_handle = []
        self.state = SphinxStripperState.SEARCHING

    def process_line(self, line: str) -> None:
        if self.state == SphinxStripperState.SEARCHING:
            self.search_for_directives(line)

        else:
            if self.state == SphinxStripperState.COLLECTING:
                # Assume that indented text below the directive is associated
                # with the directive.
                if line.strip() == '' or line[0] in (' ', '\t'):
                    self.collect_lines(line)
                # When we encounter non-indented text, we're done with this
                # directive.
                else:
                    self.state = SphinxStripperState.HANDLING

            if self.state == SphinxStripperState.HANDLING:
                self.handle_lines(line)

    def result(self) -> str:
        if self.state == SphinxStripperState.COLLECTING:
            self.state = SphinxStripperState.HANDLING
            self.handle_lines()

        return '\n'.join(self.handled_lines)


class RawDescriptionSphinxStrippedHelpFormatter(
    argparse.RawDescriptionHelpFormatter
):
    """An argparse formatter that strips Sphinx directives.

    CLI command docstrings can contain Sphinx directives for rendering in docs.
    But we don't want to include those directives when printing to the terminal.
    So we strip them and, if appropriate, replace them with something better
    suited to terminal output.
    """

    def _reformat(self, text: str) -> str:
        """Given a block of text, replace Sphinx directives.

        Directive handlers will be provided with the directive name, its tag,
        and all of the associated lines of text. "Association" is determined by
        those lines being indented to any degree under the directive.

        Unhandled directives will only have the directive line removed.
        """
        sphinx_stripper = SphinxStripper(self)

        for line in text.splitlines():
            sphinx_stripper.process_line(line)

        # The space at the end prevents the final blank line from being stripped
        # by argparse, which provides breathing room between the text and the
        # prompt.
        return sphinx_stripper.result() + ' '

    def _format_text(self, text: str) -> str:
        # This overrides an arparse method that is not technically a public API.
        return super()._format_text(self._reformat(text))

    def _handle_directive_code_block(  # pylint: disable=no-self-use
        self, tag: str, lines: List[str]
    ) -> List[str]:
        if tag == 'bash':
            processed_lines = []

            for line in lines:
                if line.strip() == '':
                    processed_lines.append(line)
                else:
                    stripped_line = line.lstrip()
                    indent = len(line) - len(stripped_line)
                    spaces = ' ' * indent
                    processed_line = f'{spaces}$ {stripped_line}'
                    processed_lines.append(processed_line)

            return processed_lines

        return lines


class _ParserAdder(Protocol):
    """Return type for _parser_adder.

    Essentially expresses the type of __call__, which cannot be expressed in
    type annotations.
    """

    def __call__(
        self, subcommand_handler: Callable[..., None], *args: Any, **kwargs: Any
    ) -> argparse.ArgumentParser:
        ...


def _parser_adder(subcommand_parser) -> _ParserAdder:
    """Create subcommand parsers with a consistent format.

    When given a subcommand handler, this will produce a parser that pulls the
    description, help, and epilog values from its docstring, and passes parsed
    args on to to the function.

    Create a subcommand parser, then feed it to this to get an `add_parser`
    function:

    .. code-block:: python

        subcommand_parser = parser_root.add_subparsers(help='Subcommands')
        add_parser = _parser_adder(subcommand_parser)

    Then use `add_parser` instead of `subcommand_parser.add_parser`.
    """

    def _add_parser(
        subcommand_handler: Callable[..., None], *args, **kwargs
    ) -> argparse.ArgumentParser:
        doc = _ParsedDocstring(subcommand_handler)
        default_kwargs = dict(
            # Displayed in list of subcommands
            description=doc.description,
            # Displayed as top-line summary for this subcommand's help
            help=doc.description,
            # Displayed as detailed help text for this subcommand's help
            epilog=doc.epilog,
            # Ensures that formatting is preserved and Sphinx directives are
            # stripped out when printing to the terminal
            formatter_class=RawDescriptionSphinxStrippedHelpFormatter,
        )

        new_kwargs = {**default_kwargs, **kwargs}
        parser = subcommand_parser.add_parser(*args, **new_kwargs)
        parser.set_defaults(func=subcommand_handler)
        return parser

    return _add_parser


def _build_argument_parser() -> argparse.ArgumentParser:
    parser_root = argparse.ArgumentParser(prog='pw ide', description=__doc__)

    parser_root.set_defaults(
        func=lambda *_args, **_kwargs: parser_root.print_help()
    )

    subcommand_parser = parser_root.add_subparsers(help='Subcommands')
    add_parser = _parser_adder(subcommand_parser)

    add_parser(cmd_setup, 'setup')

    parser_reset = add_parser(cmd_reset, 'reset')
    parser_reset.add_argument(
        '--hard',
        action='store_true',
        help='completely remove the .pw_ide working '
        'dir and supported editor files',
    )

    parser_cpp = add_parser(cmd_cpp, 'cpp')
    parser_cpp.add_argument(
        '-l',
        '--list',
        dest='should_list_targets',
        action='store_true',
        help='list the targets available for C/C++ ' 'language analysis',
    )
    parser_cpp.add_argument(
        '-g',
        '--get',
        dest='should_get_target',
        action='store_true',
        help='print the current target used for C/C++ ' 'language analysis',
    )
    parser_cpp.add_argument(
        '-s',
        '--set',
        dest='target_to_set',
        metavar='TARGET',
        help='set the target to use for C/C++ language ' 'server analysis',
    )
    parser_cpp.add_argument(
        '--set-default',
        dest='use_default_target',
        action='store_true',
        help='set the C/C++ analysis target to the default '
        'defined in pw_ide settings',
    )
    parser_cpp.add_argument(
        '--no-override',
        dest='override_current_target',
        action='store_const',
        const=False,
        default=True,
        help='if called with --set, don\'t override the '
        'current target if one is already set',
    )
    parser_cpp.add_argument(
        '--ninja',
        dest='should_run_ninja',
        action='store_true',
        help='use Ninja to generate a compilation database',
    )
    parser_cpp.add_argument(
        '--gn',
        dest='should_run_gn',
        action='store_true',
        help='run gn gen {out} --export-compile-commands, '
        'along with any other arguments defined in args.gn',
    )
    parser_cpp.add_argument(
        '-p',
        '--process',
        dest='compdb_file_paths',
        metavar='COMPILATION_DATABASE_FILES',
        type=Path,
        nargs='*',
        help='process a file or several files matching '
        'the clang compilation database format',
    )
    parser_cpp.add_argument(
        '--build-dir',
        type=Path,
        help='override the build directory defined in ' 'pw_ide settings',
    )
    parser_cpp.add_argument(
        '--clangd-command',
        action='store_true',
        help='print the command for your system that runs '
        'clangd in the activated Pigweed environment',
    )
    parser_cpp.add_argument(
        '--clangd-command-for',
        dest='clangd_command_system',
        metavar='SYSTEM',
        help='print the command for the specified system '
        'that runs clangd in the activated Pigweed '
        'environment',
    )

    parser_python = add_parser(cmd_python, 'python')
    parser_python.add_argument(
        '--venv',
        dest='should_print_venv',
        action='store_true',
        help='print the path to the Pigweed Python ' 'virtual environment',
    )

    parser_vscode = add_parser(cmd_vscode, 'vscode')
    parser_vscode.add_argument(
        '--include',
        nargs='+',
        type=VscSettingsType,
        metavar='SETTINGS_TYPE',
        help='update only these settings types',
    )
    parser_vscode.add_argument(
        '--exclude',
        nargs='+',
        type=VscSettingsType,
        metavar='SETTINGS_TYPE',
        help='do not update these settings types',
    )
    parser_vscode.add_argument(
        '--no-override',
        action='store_true',
        help='don\'t overwrite existing active ' 'settings files',
    )

    parser_clear = add_parser(cmd_clear, 'clear')
    parser_clear.add_argument(
        '--compdb',
        action='store_true',
        help='delete all compilation database from ' 'the working directory',
    )
    parser_clear.add_argument(
        '--cache',
        action='store_true',
        help='delete all compilation database caches '
        'from the working directory',
    )
    parser_clear.add_argument(
        '--editor',
        metavar='EDITOR',
        help='delete the active settings file for '
        'the provided supported editor',
    )
    parser_clear.add_argument(
        '--editor-backups',
        metavar='EDITOR',
        help='delete backup settings files for '
        'the provided supported editor',
    )

    return parser_root


def _parse_args() -> argparse.Namespace:
    args = _build_argument_parser().parse_args()
    return args


def _dispatch_command(func: Callable, **kwargs: Dict[str, Any]) -> int:
    """Dispatch arguments to a subcommand handler.

    Each CLI subcommand is handled by handler function, which is registered
    with the subcommand parser with `parser.set_defaults(func=handler)`.
    By calling this function with the parsed args, the appropriate subcommand
    handler is called, and the arguments are passed to it as kwargs.
    """
    return func(**kwargs)


def parse_args_and_dispatch_command() -> int:
    return _dispatch_command(**vars(_parse_args()))



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Serializes an Environment into a shell file."""

import inspect

# Disable super() warnings since this file must be Python 2 compatible.
# pylint: disable=super-with-arguments


class _BaseShellVisitor(object):  # pylint: disable=useless-object-inheritance
    def __init__(self, *args, **kwargs):
        pathsep = kwargs.pop('pathsep', ':')
        super(_BaseShellVisitor, self).__init__(*args, **kwargs)
        self._pathsep = pathsep
        self._outs = None

    def _remove_value_from_path(self, variable, value):
        return (
            '{variable}="$(echo "${variable}"'
            ' | sed "s|{pathsep}{value}{pathsep}|{pathsep}|g;"'
            ' | sed "s|^{value}{pathsep}||g;"'
            ' | sed "s|{pathsep}{value}$||g;"'
            ')"\nexport {variable}\n'.format(
                variable=variable, value=value, pathsep=self._pathsep
            )
        )

    def visit_hash(self, hash):  # pylint: disable=redefined-builtin
        del hash
        self._outs.write(
            inspect.cleandoc(
                '''
        # This should detect bash and zsh, which have a hash command that must
        # be called to get it to forget past commands. Without forgetting past
        # commands the $PATH changes we made may not be respected.
        if [ -n "${BASH:-}" -o -n "${ZSH_VERSION:-}" ] ; then
            hash -r\n
        fi
        '''
            )
        )


class ShellVisitor(_BaseShellVisitor):
    """Serializes an Environment into a shell file."""

    def __init__(self, *args, **kwargs):
        super(ShellVisitor, self).__init__(*args, **kwargs)
        self._replacements = ()

    def serialize(self, env, outs):
        try:
            self._replacements = tuple(
                (key, env.get(key) if value is None else value)
                for key, value in env.replacements
            )
            self._outs = outs

            env.accept(self)

        finally:
            self._replacements = ()
            self._outs = None

    def _apply_replacements(self, action):
        value = action.value
        for var, replacement in self._replacements:
            if var != action.name:
                value = value.replace(replacement, '${}'.format(var))
        return value

    def visit_set(self, set):  # pylint: disable=redefined-builtin
        value = self._apply_replacements(set)
        self._outs.write(
            '{name}="{value}"\nexport {name}\n'.format(
                name=set.name, value=value
            )
        )

    def visit_clear(self, clear):
        self._outs.write('unset {name}\n'.format(**vars(clear)))

    def visit_remove(self, remove):
        value = self._apply_replacements(remove)
        self._outs.write(
            '# Remove \n#   {value}\n# from {name} before adding it '
            'back.\n'.format(value=remove.value, name=remove.name)
        )
        self._outs.write(self._remove_value_from_path(remove.name, value))

    def _join(self, *args):
        if len(args) == 1 and isinstance(args[0], (list, tuple)):
            args = args[0]
        return self._pathsep.join(args)

    def visit_prepend(self, prepend):
        value = self._apply_replacements(prepend)
        value = self._join(value, '${}'.format(prepend.name))
        self._outs.write(
            '{name}="{value}"\nexport {name}\n'.format(
                name=prepend.name, value=value
            )
        )

    def visit_append(self, append):
        value = self._apply_replacements(append)
        value = self._join('${}'.format(append.name), value)
        self._outs.write(
            '{name}="{value}"\nexport {name}\n'.format(
                name=append.name, value=value
            )
        )

    def visit_echo(self, echo):
        # TODO(mohrr) use shlex.quote().
        self._outs.write('if [ -z "${PW_ENVSETUP_QUIET:-}" ]; then\n')
        if echo.newline:
            self._outs.write('  echo "{}"\n'.format(echo.value))
        else:
            self._outs.write('  echo -n "{}"\n'.format(echo.value))
        self._outs.write('fi\n')

    def visit_comment(self, comment):
        for line in comment.value.splitlines():
            self._outs.write('# {}\n'.format(line))

    def visit_command(self, command):
        # TODO(mohrr) use shlex.quote here?
        self._outs.write('{}\n'.format(' '.join(command.command)))
        if not command.exit_on_error:
            return

        # Assume failing command produced relevant output.
        self._outs.write('if [ "$?" -ne 0 ]; then\n  return 1\nfi\n')

    def visit_doctor(self, doctor):
        self._outs.write('if [ -z "$PW_ACTIVATE_SKIP_CHECKS" ]; then\n')
        self.visit_command(doctor)
        self._outs.write('else\n')
        self._outs.write(
            'echo Skipping environment check because '
            'PW_ACTIVATE_SKIP_CHECKS is set\n'
        )
        self._outs.write('fi\n')

    def visit_blank_line(self, blank_line):
        del blank_line
        self._outs.write('\n')

    def visit_function(self, function):
        self._outs.write(
            '{name}() {{\n{body}\n}}\n'.format(
                name=function.name, body=function.body
            )
        )


class DeactivateShellVisitor(_BaseShellVisitor):
    """Removes values from an Environment."""

    def __init__(self, *args, **kwargs):
        pathsep = kwargs.pop('pathsep', ':')
        super(DeactivateShellVisitor, self).__init__(*args, **kwargs)
        self._pathsep = pathsep

    def serialize(self, env, outs):
        try:
            self._outs = outs

            env.accept(self)

        finally:
            self._outs = None

    def visit_set(self, set):  # pylint: disable=redefined-builtin
        if set.deactivate:
            self._outs.write('unset {name}\n'.format(name=set.name))

    def visit_clear(self, clear):
        pass  # Not relevant.

    def visit_remove(self, remove):
        pass  # Not relevant.

    def visit_prepend(self, prepend):
        self._outs.write(
            self._remove_value_from_path(prepend.name, prepend.value)
        )

    def visit_append(self, append):
        self._outs.write(
            self._remove_value_from_path(append.name, append.value)
        )

    def visit_echo(self, echo):
        pass  # Not relevant.

    def visit_comment(self, comment):
        pass  # Not relevant.

    def visit_command(self, command):
        pass  # Not relevant.

    def visit_doctor(self, doctor):
        pass  # Not relevant.

    def visit_blank_line(self, blank_line):
        pass  # Not relevant.

    def visit_function(self, function):
        pass  # Not relevant.



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Parses the arguments in a Base64-encoded tokenized message.

This is useful for attempting to decode tokenized messages with arguments for
which the token is not recognized.
"""

import argparse
import base64
from dataclasses import dataclass
import logging
import sys
from typing import Collection, Iterable, Iterator, Sequence

import pw_cli.log
from pw_tokenizer.decode import FormatString, FormattedString

_LOG: logging.Logger = logging.getLogger('pw_tokenizer')

DEFAULT_FORMAT_SPECS = (
    '%s',
    '%d',
    '%f',
)

DEFAULT_MAX_ARGS = 8
PREFIX = '$'


def attempt_to_decode(
    arg_data: bytes,
    format_specs: Collection[str] = DEFAULT_FORMAT_SPECS,
    max_args: int = DEFAULT_MAX_ARGS,
    yield_failures: bool = False,
) -> Iterator[FormattedString]:
    """Attempts to decode arguments using the provided format specifiers."""
    format_strings = [(0, '')]  # (argument count, format string)

    # Each argument requires at least 1 byte.
    max_args = min(max_args, len(arg_data))

    while format_strings:
        arg_count, string = format_strings.pop(0)
        decode_attempt = FormatString(string).format(arg_data)

        if yield_failures or decode_attempt.ok():
            yield decode_attempt

        if arg_count < max_args:
            format_strings.extend(
                (arg_count + 1, string + spec) for spec in format_specs
            )


@dataclass(frozen=True)
class TokenizedMessage:
    string: str
    binary: bytes

    @property
    def token(self) -> int:
        return int.from_bytes(self.binary[:4], 'little')

    @property
    def binary_args(self) -> bytes:
        return self.binary[4:]

    @classmethod
    def parse(cls, message: str, prefix: str = '$') -> 'TokenizedMessage':
        if not message.startswith(prefix):
            raise ValueError(
                f'{message} does not start with {prefix!r} as expected'
            )

        binary = base64.b64decode(message[1:])

        if len(binary) < 4:
            raise ValueError(
                f'{message} is only {len(binary)} bytes; '
                'tokenized messages must be at least 4 bytes'
            )

        return cls(message, binary)


def _read_stdin():
    try:
        while True:
            yield input()
    except KeyboardInterrupt:
        return


def _text_list(items: Sequence, conjunction: str = 'or') -> str:
    if len(items) == 1:
        return str(items[0])

    return f'{", ".join(str(i) for i in items[:-1])} {conjunction} {items[-1]}'


def main(
    messages: Iterable[str],
    max_args: int,
    specs: Sequence[str],
    show_failures: bool,
) -> int:
    """Parses the arguments for a series of tokenized messages."""
    exit_code = 0

    for message in iter(messages) if messages else _read_stdin():
        if not message:
            continue

        if not message.startswith(PREFIX):
            message = PREFIX + message

        _LOG.info('Decoding arguments for %r', message)
        try:
            parsed = TokenizedMessage.parse(message)
        except ValueError as exc:
            _LOG.error('%s', exc)
            exit_code = 2
            continue

        _LOG.info(
            'Binary: %r [%s] (%d bytes)',
            parsed.binary,
            parsed.binary.hex(' ', 1),
            len(parsed.binary),
        )
        _LOG.info('Token:  0x%08x', parsed.token)
        _LOG.info(
            'Args:   %r [%s] (%d bytes)',
            parsed.binary_args,
            parsed.binary_args.hex(' ', 1),
            len(parsed.binary_args),
        )
        _LOG.info(
            'Decoding with up to %d %s arguments', max_args, _text_list(specs)
        )

        results = sorted(
            attempt_to_decode(
                parsed.binary_args, specs, max_args, show_failures
            ),
            key=FormattedString.score,
            reverse=True,
        )

        if not any(result.ok() for result in results):
            _LOG.warning(
                '  No combinations of up to %d %s arguments decoded '
                'successfully',
                max_args,
                _text_list(specs),
            )
            exit_code = 1

        for i, result in enumerate(results, 1):
            _LOG.info(  # pylint: disable=logging-fstring-interpolation
                f'  Attempt %{len(str(len(results)))}d: [%s] %s',
                i,
                ' '.join(str(a.specifier) for a in result.args),
                ' '.join(str(a) for a in result.args),
            )
        print()

    return exit_code


def _parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        '--max-args',
        default=DEFAULT_MAX_ARGS,
        type=int,
        help='Maximum number of printf-style arguments',
    )
    parser.add_argument(
        '--specs',
        nargs='*',
        default=DEFAULT_FORMAT_SPECS,
        help='Which printf-style format specifiers to check',
    )
    parser.add_argument(
        '--show-failures',
        action='store_true',
        help='Show argument combintations that fail to decode',
    )
    parser.add_argument(
        'messages',
        nargs='*',
        help=(
            'Base64-encoded tokenized messages to decode; omit to read from '
            'stdin'
        ),
    )
    return parser.parse_args()


if __name__ == '__main__':
    pw_cli.log.install()
    sys.exit(main(**vars(_parse_args())))



# Copyright 2020 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Stores the environment changes necessary for Pigweed."""

import contextlib
import os
import re

# The order here is important. On Python 2 we want StringIO.StringIO and not
# io.StringIO. On Python 3 there is no StringIO module so we want io.StringIO.
# Not using six because six is not a standard package we can expect to have
# installed in the system Python.
try:
    from StringIO import StringIO  # type: ignore
except ImportError:
    from io import StringIO

from . import apply_visitor
from . import batch_visitor
from . import gni_visitor
from . import json_visitor
from . import shell_visitor

# Disable super() warnings since this file must be Python 2 compatible.
# pylint: disable=super-with-arguments


class BadNameType(TypeError):
    pass


class BadValueType(TypeError):
    pass


class EmptyValue(ValueError):
    pass


class NewlineInValue(TypeError):
    pass


class BadVariableName(ValueError):
    pass


class UnexpectedAction(ValueError):
    pass


class AcceptNotOverridden(TypeError):
    pass


class _Action(object):  # pylint: disable=useless-object-inheritance
    def unapply(self, env, orig_env):
        pass

    def accept(self, visitor):
        del visitor
        raise AcceptNotOverridden(
            'accept() not overridden for {}'.format(self.__class__.__name__)
        )

    def write_deactivate(
        self, outs, windows=(os.name == 'nt'), replacements=()
    ):
        pass


class _VariableAction(_Action):
    # pylint: disable=keyword-arg-before-vararg
    def __init__(self, name, value, allow_empty_values=False, *args, **kwargs):
        super(_VariableAction, self).__init__(*args, **kwargs)
        self.name = name
        self.value = value
        self.allow_empty_values = allow_empty_values

        self._check()

    def _check(self):
        try:
            # In python2, unicode is a distinct type.
            valid_types = (str, unicode)
        except NameError:
            valid_types = (str,)

        if not isinstance(self.name, valid_types):
            raise BadNameType(
                'variable name {!r} not of type str'.format(self.name)
            )
        if not isinstance(self.value, valid_types):
            raise BadValueType(
                '{!r} value {!r} not of type str'.format(self.name, self.value)
            )

        # Empty strings as environment variable values have different behavior
        # on different operating systems. Just don't allow them.
        if not self.allow_empty_values and self.value == '':
            raise EmptyValue(
                '{!r} value {!r} is the empty string'.format(
                    self.name, self.value
                )
            )

        # Many tools have issues with newlines in environment variable values.
        # Just don't allow them.
        if '\n' in self.value:
            raise NewlineInValue(
                '{!r} value {!r} contains a newline'.format(
                    self.name, self.value
                )
            )

        if not re.match(r'^[A-Z_][A-Z0-9_]*$', self.name, re.IGNORECASE):
            raise BadVariableName('bad variable name {!r}'.format(self.name))

    def unapply(self, env, orig_env):
        if self.name in orig_env:
            env[self.name] = orig_env[self.name]
        else:
            env.pop(self.name, None)

    def __repr__(self):
        return '{}({}, {})'.format(
            self.__class__.__name__, self.name, self.value
        )


class Set(_VariableAction):
    """Set a variable."""

    def __init__(self, *args, **kwargs):
        deactivate = kwargs.pop('deactivate', True)
        super(Set, self).__init__(*args, **kwargs)
        self.deactivate = deactivate

    def accept(self, visitor):
        visitor.visit_set(self)


class Clear(_VariableAction):
    """Remove a variable from the environment."""

    def __init__(self, *args, **kwargs):
        kwargs['value'] = ''
        kwargs['allow_empty_values'] = True
        super(Clear, self).__init__(*args, **kwargs)

    def accept(self, visitor):
        visitor.visit_clear(self)


class Remove(_VariableAction):
    """Remove a value from a PATH-like variable."""

    def accept(self, visitor):
        visitor.visit_remove(self)


class BadVariableValue(ValueError):
    pass


def _append_prepend_check(action):
    if '=' in action.value:
        raise BadVariableValue('"{}" contains "="'.format(action.value))


class Prepend(_VariableAction):
    """Prepend a value to a PATH-like variable."""

    def __init__(self, name, value, join, *args, **kwargs):
        super(Prepend, self).__init__(name, value, *args, **kwargs)
        self._join = join

    def _check(self):
        super(Prepend, self)._check()
        _append_prepend_check(self)

    def accept(self, visitor):
        visitor.visit_prepend(self)


class Append(_VariableAction):
    """Append a value to a PATH-like variable. (Uncommon, see Prepend.)"""

    def __init__(self, name, value, join, *args, **kwargs):
        super(Append, self).__init__(name, value, *args, **kwargs)
        self._join = join

    def _check(self):
        super(Append, self)._check()
        _append_prepend_check(self)

    def accept(self, visitor):
        visitor.visit_append(self)


class BadEchoValue(ValueError):
    pass


class Echo(_Action):
    """Echo a value to the terminal."""

    def __init__(self, value, newline, *args, **kwargs):
        # These values act funny on Windows.
        if value.lower() in ('off', 'on'):
            raise BadEchoValue(value)
        super(Echo, self).__init__(*args, **kwargs)
        self.value = value
        self.newline = newline

    def accept(self, visitor):
        visitor.visit_echo(self)

    def __repr__(self):
        return 'Echo({}, newline={})'.format(self.value, self.newline)


class Comment(_Action):
    """Add a comment to the init script."""

    def __init__(self, value, *args, **kwargs):
        super(Comment, self).__init__(*args, **kwargs)
        self.value = value

    def accept(self, visitor):
        visitor.visit_comment(self)

    def __repr__(self):
        return 'Comment({})'.format(self.value)


class Command(_Action):
    """Run a command."""

    def __init__(self, command, *args, **kwargs):
        exit_on_error = kwargs.pop('exit_on_error', True)
        super(Command, self).__init__(*args, **kwargs)
        assert isinstance(command, (list, tuple))
        self.command = command
        self.exit_on_error = exit_on_error

    def accept(self, visitor):
        visitor.visit_command(self)

    def __repr__(self):
        return 'Command({})'.format(self.command)


class Doctor(Command):
    def __init__(self, *args, **kwargs):
        log_level = 'warn' if 'PW_ENVSETUP_QUIET' in os.environ else 'info'
        cmd = ['pw', '--no-banner', '--loglevel', log_level, 'doctor']
        super(Doctor, self).__init__(command=cmd, *args, **kwargs)

    def accept(self, visitor):
        visitor.visit_doctor(self)

    def __repr__(self):
        return 'Doctor()'


class BlankLine(_Action):
    """Write a blank line to the init script."""

    def accept(self, visitor):
        visitor.visit_blank_line(self)

    def __repr__(self):
        return 'BlankLine()'


class Function(_Action):
    def __init__(self, name, body, *args, **kwargs):
        super(Function, self).__init__(*args, **kwargs)
        self.name = name
        self.body = body

    def accept(self, visitor):
        visitor.visit_function(self)

    def __repr__(self):
        return 'Function({}, {})'.format(self.name, self.body)


class Hash(_Action):
    def accept(self, visitor):
        visitor.visit_hash(self)

    def __repr__(self):
        return 'Hash()'


class Join(object):  # pylint: disable=useless-object-inheritance
    def __init__(self, pathsep=os.pathsep):
        self.pathsep = pathsep


# TODO(mohrr) remove disable=useless-object-inheritance once in Python 3.
# pylint: disable=useless-object-inheritance
class Environment(object):
    """Stores the environment changes necessary for Pigweed.

    These changes can be accessed by writing them to a file for bash-like
    shells to source or by using this as a context manager.
    """

    def __init__(self, *args, **kwargs):
        pathsep = kwargs.pop('pathsep', os.pathsep)
        windows = kwargs.pop('windows', os.name == 'nt')
        allcaps = kwargs.pop('allcaps', windows)
        super(Environment, self).__init__(*args, **kwargs)
        self._actions = []
        self._pathsep = pathsep
        self._windows = windows
        self._allcaps = allcaps
        self.replacements = []
        self._join = Join(pathsep)
        self._finalized = False

    def add_replacement(self, variable, value=None):
        self.replacements.append((variable, value))

    def normalize_key(self, name):
        if self._allcaps:
            try:
                return name.upper()
            except AttributeError:
                # The _Action class has code to handle incorrect types, so
                # we just ignore this error here.
                pass
        return name

    # A newline is printed after each high-level operation. Top-level
    # operations should not invoke each other (this is why _remove() exists).

    def set(self, name, value, deactivate=True):
        """Set a variable."""
        assert not self._finalized
        name = self.normalize_key(name)
        self._actions.append(Set(name, value, deactivate=deactivate))
        self._blankline()

    def clear(self, name):
        """Remove a variable."""
        assert not self._finalized
        name = self.normalize_key(name)
        self._actions.append(Clear(name))
        self._blankline()

    def _remove(self, name, value):
        """Remove a value from a variable."""
        assert not self._finalized
        name = self.normalize_key(name)
        if self.get(name, None):
            self._actions.append(Remove(name, value, self._pathsep))

    def remove(self, name, value):
        """Remove a value from a PATH-like variable."""
        assert not self._finalized
        self._remove(name, value)
        self._blankline()

    def append(self, name, value):
        """Add a value to a PATH-like variable. Rarely used, see prepend()."""
        assert not self._finalized
        name = self.normalize_key(name)
        if self.get(name, None):
            self._remove(name, value)
            self._actions.append(Append(name, value, self._join))
        else:
            self._actions.append(Set(name, value))
        self._blankline()

    def prepend(self, name, value):
        """Add a value to the beginning of a PATH-like variable."""
        assert not self._finalized
        name = self.normalize_key(name)
        if self.get(name, None):
            self._remove(name, value)
            self._actions.append(Prepend(name, value, self._join))
        else:
            self._actions.append(Set(name, value))
        self._blankline()

    def echo(self, value='', newline=True):
        """Echo a value to the terminal."""
        # echo() deliberately ignores self._finalized.
        self._actions.append(Echo(value, newline))
        if value:
            self._blankline()

    def comment(self, comment):
        """Add a comment to the init script."""
        # comment() deliberately ignores self._finalized.
        self._actions.append(Comment(comment))
        self._blankline()

    def command(self, command, exit_on_error=True):
        """Run a command."""
        # command() deliberately ignores self._finalized.
        self._actions.append(Command(command, exit_on_error=exit_on_error))
        self._blankline()

    def doctor(self):
        """Run 'pw doctor'."""
        self._actions.append(Doctor())

    def function(self, name, body):
        """Define a function."""
        assert not self._finalized
        self._actions.append(Command(name, body))
        self._blankline()

    def _blankline(self):
        self._actions.append(BlankLine())

    def finalize(self):
        """Run cleanup at the end of environment setup."""
        assert not self._finalized
        self._finalized = True
        self._actions.append(Hash())
        self._blankline()

        if not self._windows:
            buf = StringIO()
            self.write_deactivate(buf)
            self._actions.append(Function('_pw_deactivate', buf.getvalue()))
            self._blankline()

    def accept(self, visitor):
        for action in self._actions:
            action.accept(visitor)

    def gni(self, outs, project_root):
        gni_visitor.GNIVisitor(project_root).serialize(self, outs)

    def json(self, outs):
        json_visitor.JSONVisitor().serialize(self, outs)

    def write(self, outs):
        if self._windows:
            visitor = batch_visitor.BatchVisitor(pathsep=self._pathsep)
        else:
            visitor = shell_visitor.ShellVisitor(pathsep=self._pathsep)
        visitor.serialize(self, outs)

    def write_deactivate(self, outs):
        if self._windows:
            return
        visitor = shell_visitor.DeactivateShellVisitor(pathsep=self._pathsep)
        visitor.serialize(self, outs)

    @contextlib.contextmanager
    def __call__(self, export=True):
        """Set environment as if this was written to a file and sourced.

        Within this context os.environ is updated with the environment
        defined by this object. If export is False, os.environ is not updated,
        but in both cases the updated environment is yielded.

        On exit, previous environment is restored. See contextlib documentation
        for details on how this function is structured.

        Args:
          export(bool): modify the environment of the running process (and
            thus, its subprocesses)

        Yields the new environment object.
        """
        orig_env = {}
        try:
            if export:
                orig_env = os.environ.copy()
                env = os.environ
            else:
                env = os.environ.copy()

            apply = apply_visitor.ApplyVisitor(pathsep=self._pathsep)
            apply.apply(self, env)

            yield env

        finally:
            if export:
                for key in set(os.environ):
                    try:
                        os.environ[key] = orig_env[key]
                    except KeyError:
                        del os.environ[key]
                for key in set(orig_env) - set(os.environ):
                    os.environ[key] = orig_env[key]

    def get(self, key, default=None):
        """Get the value of a variable within context of this object."""
        key = self.normalize_key(key)
        with self(export=False) as env:
            return env.get(key, default)

    def __getitem__(self, key):
        """Get the value of a variable within context of this object."""
        key = self.normalize_key(key)
        with self(export=False) as env:
            return env[key]



# Copyright 2022 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""CommandRunner dialog classes."""

from __future__ import annotations
import functools
import logging
import re
from typing import (
    Callable,
    Iterable,
    Iterator,
    List,
    Optional,
    TYPE_CHECKING,
    Tuple,
)

from prompt_toolkit.buffer import Buffer
from prompt_toolkit.filters import Condition
from prompt_toolkit.formatted_text import StyleAndTextTuples
from prompt_toolkit.formatted_text.utils import fragment_list_to_text
from prompt_toolkit.layout.utils import explode_text_fragments
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.key_binding import (
    KeyBindings,
    KeyBindingsBase,
    KeyPressEvent,
)
from prompt_toolkit.layout import (
    AnyContainer,
    ConditionalContainer,
    DynamicContainer,
    FormattedTextControl,
    HSplit,
    VSplit,
    Window,
    WindowAlign,
)
from prompt_toolkit.widgets import MenuItem
from prompt_toolkit.widgets import TextArea

from pw_console.widgets import (
    create_border,
    mouse_handlers,
    to_keybind_indicator,
)

if TYPE_CHECKING:
    from pw_console.console_app import ConsoleApp

_LOG = logging.getLogger(__package__)


def flatten_menu_items(
    items: List[MenuItem], prefix: str = ''
) -> Iterator[Tuple[str, Callable]]:
    """Flatten nested prompt_toolkit MenuItems into text and callable tuples."""
    for item in items:
        new_text = []
        if prefix:
            new_text.append(prefix)
        new_text.append(item.text)
        new_prefix = ' > '.join(new_text)

        if item.children:
            yield from flatten_menu_items(item.children, new_prefix)
        elif item.handler:
            # Skip this item if it's a separator or disabled.
            if item.text == '-' or item.disabled:
                continue
            yield (new_prefix, item.handler)


def highlight_matches(
    regexes: Iterable[re.Pattern], line_fragments: StyleAndTextTuples
) -> StyleAndTextTuples:
    """Highlight regex matches in prompt_toolkit FormattedTextTuples."""
    line_text = fragment_list_to_text(line_fragments)
    exploded_fragments = explode_text_fragments(line_fragments)

    def apply_highlighting(
        fragments: StyleAndTextTuples, index: int, matching_regex_index: int = 0
    ) -> None:
        # Expand all fragments and apply the highlighting style.
        old_style, _text, *_ = fragments[index]
        # There are 6 fuzzy-highlight styles defined in style.py. Get an index
        # from 0-5 to use one style after the other in turn.
        style_index = matching_regex_index % 6
        fragments[index] = (
            old_style + f' class:command-runner-fuzzy-highlight-{style_index} ',
            fragments[index][1],
        )

    # Highlight each non-overlapping search match.
    for regex_i, regex in enumerate(regexes):
        for match in regex.finditer(line_text):
            for fragment_i in range(match.start(), match.end()):
                apply_highlighting(exploded_fragments, fragment_i, regex_i)

    return exploded_fragments


class CommandRunner:
    """CommandRunner dialog box."""

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        application: ConsoleApp,
        window_title: Optional[str] = None,
        load_completions: Optional[
            Callable[[], List[Tuple[str, Callable]]]
        ] = None,
        width: int = 80,
        height: int = 10,
    ):
        # Parent pw_console application
        self.application = application
        # Visibility toggle
        self.show_dialog = False
        # Tracks the last focused container, to enable restoring focus after
        # closing the dialog.
        self.last_focused_pane = None

        # List of all possible completion items
        self.completions: List[Tuple[str, Callable]] = []
        # Formatted text fragments of matched items
        self.completion_fragments: List[StyleAndTextTuples] = []

        # Current selected item tracking variables
        self.selected_item: int = 0
        self.selected_item_text: str = ''
        self.selected_item_handler: Optional[Callable] = None
        # Previous input text
        self.last_input_field_text: str = 'EMPTY'
        # Previous selected item
        self.last_selected_item: int = 0

        # Dialog width, height and title
        self.width = width
        self.height = height
        self.window_title: str

        # Callable to fetch completion items
        self.load_completions: Callable[[], List[Tuple[str, Callable]]]

        # Command runner text input field
        self.input_field = TextArea(
            prompt=[
                (
                    'class:command-runner-setting',
                    '> ',
                    functools.partial(
                        mouse_handlers.on_click,
                        self.focus_self,
                    ),
                )
            ],
            focusable=True,
            focus_on_click=True,
            scrollbar=False,
            multiline=False,
            height=1,
            dont_extend_height=True,
            dont_extend_width=False,
            accept_handler=self._command_accept_handler,
            history=InMemoryHistory(),
        )
        # Set additional keybindings for the input field
        self.input_field.control.key_bindings = self._create_key_bindings()

        # Container for the Cancel and Run buttons
        input_field_buttons_container = ConditionalContainer(
            Window(
                content=FormattedTextControl(
                    self._get_input_field_button_fragments,
                    focusable=False,
                    show_cursor=False,
                ),
                height=1,
                align=WindowAlign.RIGHT,
                dont_extend_width=True,
            ),
            filter=Condition(lambda: self.content_width() > 40),
        )

        # Container for completion matches
        command_items_window = Window(
            content=FormattedTextControl(
                self.render_completion_items,
                show_cursor=False,
                focusable=False,
            ),
            align=WindowAlign.LEFT,
            dont_extend_width=False,
            height=self.height,
        )

        # Main content HSplit
        self.command_runner_content = HSplit(
            [
                # Input field and buttons on the same line
                VSplit(
                    [
                        self.input_field,
                        input_field_buttons_container,
                    ]
                ),
                # Completion items below
                command_items_window,
            ],
            style='class:command-runner class:theme-fg-default',
        )

        # Set completions if passed in.
        self.set_completions(window_title, load_completions)

        # bordered_content wraps the above command_runner_content in a border.
        self.bordered_content: AnyContainer
        # Root prompt_toolkit container
        self.container = ConditionalContainer(
            DynamicContainer(lambda: self.bordered_content),
            filter=Condition(lambda: self.show_dialog),
        )

    def _create_bordered_content(self) -> None:
        """Wrap self.command_runner_content in a border."""
        # This should be called whenever the window_title changes.
        self.bordered_content = create_border(
            self.command_runner_content,
            title=self.window_title,
            border_style='class:command-runner-border',
            left_margin_columns=1,
            right_margin_columns=1,
        )

    def __pt_container__(self) -> AnyContainer:
        """Return the prompt_toolkit root container for this dialog."""
        return self.container

    def _create_key_bindings(self) -> KeyBindingsBase:
        """Create additional key bindings for the command input field."""
        key_bindings = KeyBindings()
        register = self.application.prefs.register_keybinding

        @register('command-runner.cancel', key_bindings)
        def _cancel(_event: KeyPressEvent) -> None:
            """Clear input or close command."""
            if self._get_input_field_text() != '':
                self._reset_selected_item()
                return

            self.close_dialog()

        @register('command-runner.select-previous-item', key_bindings)
        def _select_previous_item(_event: KeyPressEvent) -> None:
            """Select previous completion item."""
            self._previous_item()

        @register('command-runner.select-next-item', key_bindings)
        def _select_next_item(_event: KeyPressEvent) -> None:
            """Select next completion item."""
            self._next_item()

        return key_bindings

    def content_width(self) -> int:
        """Return the smaller value of self.width and the available width."""
        window_manager_width = (
            self.application.window_manager.current_window_manager_width
        )
        if not window_manager_width:
            window_manager_width = self.width
        return min(self.width, window_manager_width)

    def focus_self(self) -> None:
        self.application.layout.focus(self)

    def close_dialog(self) -> None:
        """Close command runner dialog box."""
        self.show_dialog = False
        self._reset_selected_item()

        # Restore original focus if possible.
        if self.last_focused_pane:
            self.application.focus_on_container(self.last_focused_pane)
        else:
            # Fallback to focusing on the main menu.
            self.application.focus_main_menu()

    def open_dialog(self) -> None:
        self.show_dialog = True
        self.last_focused_pane = self.application.focused_window()
        self.focus_self()
        self.application.redraw_ui()

    def set_completions(
        self,
        window_title: Optional[str] = None,
        load_completions: Optional[
            Callable[[], List[Tuple[str, Callable]]]
        ] = None,
    ) -> None:
        """Set window title and callable to fetch possible completions.

        Call this function whenever new completion items need to be loaded.
        """
        self.window_title = window_title if window_title else 'Menu Items'
        self.load_completions = (
            load_completions if load_completions else self.load_menu_items
        )
        self._reset_selected_item()

        self.completions = []
        self.completion_fragments = []

        # Load and filter completions
        self.filter_completions()

        # (Re)create the bordered content with the window_title set.
        self._create_bordered_content()

    def reload_completions(self) -> None:
        self.completions = self.load_completions()

    def load_menu_items(self) -> List[Tuple[str, Callable]]:
        # pylint: disable=no-self-use
        return list(flatten_menu_items(self.application.menu_items))

    def _get_input_field_text(self) -> str:
        return self.input_field.buffer.text

    def _make_regexes(self, input_text) -> List[re.Pattern]:
        # pylint: disable=no-self-use
        regexes: List[re.Pattern] = []
        if not input_text:
            return regexes

        text_tokens = input_text.split(' ')
        if len(text_tokens) > 0:
            regexes = [
                re.compile(re.escape(text), re.IGNORECASE)
                for text in text_tokens
            ]

        return regexes

    def _matches_orderless(self, regexes: List[re.Pattern], text) -> bool:
        """Check if all supplied regexs match the input text."""
        # pylint: disable=no-self-use
        return all(regex.search(text) for regex in regexes)

    def filter_completions(self) -> None:
        """Filter completion items if new user input detected."""
        if not self.input_text_changed() and not self.selected_item_changed():
            return

        self.reload_completions()

        input_text = self._get_input_field_text()
        self.completion_fragments = []

        regexes = self._make_regexes(input_text)
        check_match = self._matches_orderless

        i = 0
        for text, handler in self.completions:
            if not (input_text == '' or check_match(regexes, text)):
                continue
            style = ''
            if i == self.selected_item:
                style = 'class:command-runner-selected-item'
                self.selected_item_text = text
                self.selected_item_handler = handler
                text = text.ljust(self.content_width())
            fragments: StyleAndTextTuples = highlight_matches(
                regexes, [(style, text + '\n')]
            )
            self.completion_fragments.append(fragments)
            i += 1

    def input_text_changed(self) -> bool:
        """Return True if text in the input field has changed."""
        input_text = self._get_input_field_text()
        if input_text != self.last_input_field_text:
            self.last_input_field_text = input_text
            self.selected_item = 0
            return True
        return False

    def selected_item_changed(self) -> bool:
        """Check if the user pressed up or down to select a different item."""
        return self.last_selected_item != self.selected_item

    def _next_item(self) -> None:
        self.last_selected_item = self.selected_item
        self.selected_item = min(
            # Don't move past the height of the window or the length of possible
            # items.
            min(self.height, len(self.completion_fragments)) - 1,
            self.selected_item + 1,
        )
        self.application.redraw_ui()

    def _previous_item(self) -> None:
        self.last_selected_item = self.selected_item
        self.selected_item = max(0, self.selected_item - 1)
        self.application.redraw_ui()

    def _get_input_field_button_fragments(self) -> StyleAndTextTuples:
        # Mouse handlers
        focus = functools.partial(mouse_handlers.on_click, self.focus_self)
        cancel = functools.partial(mouse_handlers.on_click, self.close_dialog)
        select_item = functools.partial(
            mouse_handlers.on_click, self._run_selected_item
        )

        separator_text = ('', ' ', focus)

        # Default button style
        button_style = 'class:toolbar-button-inactive'

        fragments: StyleAndTextTuples = []

        # Cancel button
        fragments.extend(
            to_keybind_indicator(
                key='Ctrl-c',
                description='Cancel',
                mouse_handler=cancel,
                base_style=button_style,
            )
        )
        fragments.append(separator_text)

        # Run button
        fragments.extend(
            to_keybind_indicator(
                'Enter', 'Run', select_item, base_style=button_style
            )
        )
        return fragments

    def render_completion_items(self) -> StyleAndTextTuples:
        """Render completion items."""
        fragments: StyleAndTextTuples = []

        # Update completions if any state change since the last render (new text
        # entered or arrow keys pressed).
        self.filter_completions()

        for completion_item in self.completion_fragments:
            fragments.extend(completion_item)

        return fragments

    def _reset_selected_item(self) -> None:
        self.selected_item = 0
        self.last_selected_item = 0
        self.selected_item_text = ''
        self.selected_item_handler = None
        self.last_input_field_text = 'EMPTY'
        self.input_field.buffer.reset()

    def _run_selected_item(self) -> None:
        """Run the selected action."""
        if not self.selected_item_handler:
            return
        # Save the selected item handler. This is reset by self.close_dialog()
        handler = self.selected_item_handler

        # Depending on what action is run, the command runner dialog may need to
        # be closed, left open, or closed before running the selected action.
        close_dialog = True
        close_dialog_first = False

        # Actions that launch new command runners, close_dialog should not run.
        for command_text in [
            '[File] > Insert Repl Snippet',
            '[File] > Insert Repl History',
            '[File] > Open Logger',
        ]:
            if command_text in self.selected_item_text:
                close_dialog = False
                break

        # Actions that change what is in focus should be run after closing the
        # command runner dialog.
        for command_text in [
            '[File] > Games > ',
            '[View] > Focus Next Window/Tab',
            '[View] > Focus Prev Window/Tab',
            # All help menu entries open popup windows.
            '[Help] > ',
            # This focuses on a save dialog bor.
            'Save/Export a copy',
            '[Windows] > Floating ',
        ]:
            if command_text in self.selected_item_text:
                close_dialog_first = True
                break

        # Close first if needed
        if close_dialog and close_dialog_first:
            self.close_dialog()

        # Run the selected item handler
        handler()

        # If not already closed earlier.
        if close_dialog and not close_dialog_first:
            self.close_dialog()

    def _command_accept_handler(self, _buff: Buffer) -> bool:
        """Function run when pressing Enter in the command runner input box."""
        # If at least one match is available
        if len(self.completion_fragments) > 0:
            self._run_selected_item()
            # Erase input text
            return False
        # Keep input text
        return True



# Copyright 2021 The Pigweed Authors
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may not
# use this file except in compliance with the License. You may obtain a copy of
# the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Pigweed Console progress bar functions."""
from pw_console.progress_bar.progress_bar_state import TASKS_CONTEXTVAR
from pw_console.progress_bar.progress_bar_task_counter import (
    ProgressBarTaskCounter,
)

__all__ = [
    'start_progress',
    'update_progress',
]


def start_progress(task_name: str, total: int, hide_eta=False):
    progress_state = TASKS_CONTEXTVAR.get()

    progress_state.startup_progress_bar_impl()

    assert progress_state.instance is not None

    progress_state.tasks[task_name] = ProgressBarTaskCounter(
        name=task_name,
        total=total,
        prompt_toolkit_counter=progress_state.instance(
            range(total), label=task_name
        ),
    )
    ptc = progress_state.tasks[task_name].prompt_toolkit_counter
    ptc.hide_eta = hide_eta  # type: ignore


def update_progress(
    task_name: str, count=1, completed=False, canceled=False, new_total=None
):
    progress_state = TASKS_CONTEXTVAR.get()
    # The caller may not actually get canceled and will continue trying to
    # update after an interrupt.
    if task_name not in progress_state.tasks:
        return

    # Take one action
    if completed:
        progress_state.tasks[task_name].mark_completed()
    elif canceled:
        progress_state.tasks[task_name].mark_canceled()
    elif new_total:
        progress_state.tasks[task_name].set_new_total(new_total)
    else:
        progress_state.tasks[task_name].update(count)

    # Check if all tasks are complete
    if (
        progress_state.instance is not None
        and progress_state.all_tasks_complete
    ):
        if hasattr(progress_state.instance, '__exit__'):
            progress_state.instance.__exit__()
